import re
from typing import Iterable, List, Union

from ..patterns import MeshPatt, Patt, Perm


class Basis(tuple):
    """A set of classical patterns such that none is
    contained in another within the basis.
    """

    def __new__(cls, *patts: Perm) -> "Basis":
        if not patts:
            return tuple.__new__(cls, ())
        return cls._pruner(sorted(patts))

    @classmethod
    def from_string(cls, patts: str) -> "Basis":
        """Construct a Basis from a string. It can be either 0 or 1 based and
        seperated by anything."""
        return cls(*map(Perm.to_standard, re.findall(r"\d+", patts)))

    @classmethod
    def from_iterable(cls, patts: Iterable[Perm]) -> "Basis":
        """Construct a Basis from an iterable."""
        return cls(*patts)

    @classmethod
    def _pruner(cls, patts: List[Perm]) -> "Basis":
        if len(patts[0]) == 0:
            return tuple.__new__(cls, (patts[0],))
        new_basis: List[Perm] = []
        for patt in patts:
            if patt.avoids(*new_basis):
                new_basis.append(patt)
        return tuple.__new__(cls, new_basis)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and tuple.__eq__(self, other)

    def __hash__(self) -> int:
        return tuple.__hash__(self)

    def __repr__(self) -> str:
        return f"Basis({tuple.__repr__(self)})"

    def __str__(self) -> str:
        return f'{{{", ".join(str(p) for p in self)}}}'


class MeshBasis(tuple):
    """A set of patterns such that none is
    contained in another within the basis.
    """

    @staticmethod
    def is_mesh_basis(basis: Iterable[Patt]) -> bool:
        """Checks if a collection of patterns contains any non-classical ones."""
        return not (
            isinstance(basis, Perm) or all(isinstance(patt, Perm) for patt in basis)
        )

    def __new__(cls, *patts: Union[Perm, MeshPatt]) -> "MeshBasis":
        if not patts:
            return tuple.__new__(cls, ())
        return cls._pruner(
            sorted(
                patt if isinstance(patt, MeshPatt) else MeshPatt(patt, [])
                for patt in patts
            )
        )

    @classmethod
    def from_iterable(cls, patts: Iterable[Union[Perm, MeshPatt]]) -> "MeshBasis":
        """Construct a MeshBasis from an iterable."""
        return cls(*patts)

    @classmethod
    def _pruner(cls, patts: List[MeshPatt]) -> "MeshBasis":
        if len(patts[0]) == 0:
            return tuple.__new__(cls, (patts[0],))
        new_basis: List[MeshPatt] = []
        for patt in patts:
            if patt.avoids(*new_basis):
                new_basis.append(patt)
        return tuple.__new__(cls, new_basis)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and tuple.__eq__(self, other)

    def __hash__(self) -> int:
        return tuple.__hash__(self)

    def __repr__(self) -> str:
        return f"{MeshBasis}({tuple.__repr__(self)})"

    def __str__(self) -> str:
        return f'{{{", ".join(str(p) for p in self)}}}'



from collections import deque
from itertools import islice
from typing import Deque, List, Tuple

from permuta.patterns.meshpatt import MeshPatt
from permuta.patterns.patt import Patt
from permuta.patterns.perm import Perm
from permuta.permutils.groups import dihedral_group


def _is_sorted(lis: List[int]) -> bool:
    #  Return true if w is increasing, i.e., sorted.
    return all(elem == i for elem, i in zip(lis, range(len(lis))))


def _stack_sort(perm_slice: List[int]) -> List[int]:
    n = len(perm_slice)
    if n in (0, 1):
        return perm_slice
    max_i, max_v = max(enumerate(perm_slice), key=lambda pos_elem: pos_elem[1])
    # Recursively solve without largest
    if max_i == 0:
        n_lis = _stack_sort(perm_slice[1:n])
    elif max_i == n - 1:
        n_lis = _stack_sort(perm_slice[0 : n - 1])
    else:
        n_lis = _stack_sort(perm_slice[0:max_i])
        n_lis.extend(_stack_sort(perm_slice[max_i + 1 : n]))
    n_lis.append(max_v)
    return n_lis


def stack_sortable(perm: Perm) -> bool:
    """Returns true if perm is stack sortable."""
    return _is_sorted(_stack_sort(list(perm)))


def _bubble_sort(perm_slice: List[int]) -> List[int]:
    n = len(perm_slice)
    if n in (0, 1):
        return perm_slice
    max_i, max_v = max(enumerate(perm_slice), key=lambda pos_elem: pos_elem[1])
    # Recursively solve without largest
    if max_i == 0:
        n_lis = perm_slice[1:n]
    elif max_i == n - 1:
        n_lis = _bubble_sort(perm_slice[0 : n - 1])
    else:
        n_lis = _bubble_sort(perm_slice[0:max_i])
        n_lis.extend(perm_slice[max_i + 1 : n])
    n_lis.append(max_v)
    return n_lis


def bubble_sortable(perm: Perm) -> bool:
    """Returns true if perm is stack sortable."""
    return _is_sorted(_bubble_sort(list(perm)))


def _quick_sort(perm_slice: List[int]) -> List[int]:
    assert not perm_slice or set(perm_slice) == set(
        range(min(perm_slice), max(perm_slice) + 1)
    )
    n = len(perm_slice)
    if n == 0:
        return perm_slice
    maxind = -1
    # Note that perm does not need standardizing as sfp uses left to right maxima.
    for maxind in Perm(perm_slice).strong_fixed_points():
        pass
    if maxind != -1:
        lis: List[int] = (
            _quick_sort(perm_slice[:maxind])
            + [perm_slice[maxind]]
            + _quick_sort(perm_slice[maxind + 1 :])
        )
    else:
        firstval = perm_slice[0]
        lis = (
            list(filter(lambda x: x < firstval, perm_slice))
            + [perm_slice[0]]
            + list(filter(lambda x: x > firstval, perm_slice))
        )
    return lis


def quick_sortable(perm: Perm) -> bool:
    """Returns true if perm is quicksort sortable."""
    return _is_sorted(_quick_sort(list(perm)))


_BKV_PATT = Perm((1, 0))


def bkv_sortable(perm: Perm, patterns: Tuple[Patt, ...] = ()) -> bool:
    """Check if a permutation is BKV sortable.
    See:
        https://arxiv.org/pdf/1907.08142.pdf
        https://arxiv.org/pdf/2004.01812.pdf
    """
    # See
    n = len(perm)
    inp = deque(perm)
    # the right stack read from top to bottom
    # the left stack read from top to bottom
    right_stack: Deque[int] = deque([])
    left_stack: Deque[int] = deque([])
    expected = 0
    while expected < n:
        if inp:
            right_stack.appendleft(inp[0])
            if Perm.to_standard(right_stack).avoids(*patterns):
                inp.popleft()
                continue
            right_stack.popleft()

        if right_stack:
            left_stack.appendleft(right_stack[0])
            if Perm.to_standard(left_stack).avoids(_BKV_PATT):
                right_stack.popleft()
                continue
            left_stack.popleft()

        assert left_stack
        # Normally, we would gather elements from left stack but since we only care
        # about wether it sorts the permutation, we just compare it against expected.
        if expected != left_stack.popleft():
            return False
        expected += 1
    return True


def west_2_stack_sortable(perm: Perm) -> bool:
    """Returns true if perm can be sorted by two passes through a stack"""
    return _is_sorted(_stack_sort(_stack_sort(list(perm))))


def west_3_stack_sortable(perm: Perm) -> bool:
    """Returns true if perm can be sorted by three passes through a stack"""
    return _is_sorted(_stack_sort(_stack_sort(_stack_sort(list(perm)))))


_SMOOTH_PATT = (Perm((0, 2, 1, 3)), Perm((1, 0, 3, 2)))


def smooth(perm: Perm) -> bool:
    """Returns true if the perm is smooth, i.e. 0213- and 1032-avoiding."""
    return perm.avoids(*_SMOOTH_PATT)


_FOREST_LIKE_PATT = (Perm((0, 2, 1, 3)), MeshPatt(Perm((1, 0, 3, 2)), [(2, 2)]))


def forest_like(perm: Perm) -> bool:
    """Returns true if the perm is forest like."""
    return perm.avoids(*_FOREST_LIKE_PATT)


_BAXTER_PATT = (
    MeshPatt(Perm((1, 3, 0, 2)), [(2, 2)]),
    MeshPatt(Perm((2, 0, 3, 1)), [(2, 2)]),
)


def baxter(perm: Perm) -> bool:
    """Returns true if the perm is a baxter permutation."""
    return perm.avoids(*_BAXTER_PATT)


_SIMSUN_PATT = MeshPatt(Perm((2, 1, 0)), [(1, 0), (1, 1), (2, 2)])


def simsun(perm: Perm) -> bool:
    """Returns true if the perm is a simsun permutation."""
    return perm.avoids(*_SIMSUN_PATT)


def dihedral(perm: Perm) -> bool:
    """Does perm belong to a dihedral group? We use the convention that D1 and D2 are
    not subgrroups of S1 and S2, respectively."""
    return any(perm == d_perm for d_perm in dihedral_group(len(perm)))


def in_alternating_group(perm: Perm) -> bool:
    """Does perm belong to alternating group? We use the convention that D1 and D2 are
    not subgrroups of S1 and S2, respectively."""
    n = len(perm)
    if n == 0:
        return True
    if n < 3:
        return n % 2 == 1
    return perm.count_inversions() % 2 == 0


def _perm_to_yt(perm: Perm) -> List[List[int]]:
    # transform perm to standard young table

    def insert_in_row(i, k):
        cur_row = res[i]
        found = next(((ind, cur) for ind, cur in enumerate(cur_row) if cur > k), None)
        if not found:
            cur_row.append(k)
        else:
            ind, cur = found
            cur_row[ind] = k
            if len(res) <= i + 1:
                res.append([cur])
            else:
                insert_in_row(i + 1, cur)

    res: List[List[int]] = [[perm[0]]] if perm else []
    for val in islice(perm, 1, None):
        insert_in_row(0, val)
    return res


def _tableau_contains_shape(tab: List[List[int]], shape: List[int]) -> bool:
    #  Return True if the tableaux tab contains the shape
    return len(tab) >= len(shape) and all(s <= t for s, t in zip(shape, map(len, tab)))


def yt_perm_avoids_22(perm: Perm) -> bool:
    """Returns true if perm's standard young table avoids shape [2,2]."""
    return not _tableau_contains_shape(_perm_to_yt(perm), [2, 2])


def yt_perm_avoids_32(perm: Perm) -> bool:
    """Returns true if perm's standard young table avoids shape [3,2]."""
    return not _tableau_contains_shape(_perm_to_yt(perm), [3, 2])


_AV_231_AND_MESH_PATT = (
    Perm((1, 2, 0)),
    MeshPatt(Perm((0, 1, 5, 2, 3, 4)), [(1, 6), (4, 5), (4, 6)]),
)


def av_231_and_mesh(perm: Perm) -> bool:
    """Check if perm avoids MeshPatt(Perm((0, 1, 5, 2, 3, 4)), [(1, 6), (4, 5), (4, 6)])
    and the classial pattern 231.
    """
    return perm.avoids(*_AV_231_AND_MESH_PATT)


_HARD_MESH_PATT = (
    MeshPatt(Perm((0, 1, 2)), [(0, 0), (1, 1), (2, 2), (3, 3)]),
    MeshPatt(Perm((0, 1, 2)), [(0, 3), (1, 2), (2, 1), (3, 0)]),
)


def hard_mesh(perm: Perm) -> bool:
    """Check if perm avoids MeshPatt(Perm((0, 1, 2)), [(0, 0), (1, 1), (2, 2), (3, 3)])
    and MeshPatt(Perm((0, 1, 2)), [(0, 3), (1, 2), (2, 1), (3, 0)])."""
    return perm.avoids(*_HARD_MESH_PATT)



from math import factorial
from random import randint

from permuta import MeshPatt, Perm
from permuta.bisc.perm_properties import (
    _bubble_sort,
    _is_sorted,
    _perm_to_yt,
    _quick_sort,
    _stack_sort,
    av_231_and_mesh,
    baxter,
    bkv_sortable,
    bubble_sortable,
    dihedral,
    forest_like,
    hard_mesh,
    in_alternating_group,
    quick_sortable,
    simsun,
    smooth,
    stack_sortable,
    west_2_stack_sortable,
    west_3_stack_sortable,
    yt_perm_avoids_22,
    yt_perm_avoids_32,
)


def test__is_sorted():
    assert _is_sorted([])
    assert _is_sorted([0])
    assert _is_sorted([0, 1])
    assert not _is_sorted([1, 0])
    assert _is_sorted([0, 1, 2])
    assert not _is_sorted([0, 2, 1])
    assert not _is_sorted([1, 0, 2])
    assert not _is_sorted([1, 2, 0])
    assert not _is_sorted([2, 0, 1])
    assert not _is_sorted([2, 1, 0])


def test__stack_sort():
    assert _stack_sort([]) == []
    assert _stack_sort([0]) == [0]
    assert _stack_sort([0, 1]) == [0, 1]
    assert _stack_sort([1, 0]) == [0, 1]
    assert _stack_sort([0, 1, 2]) == [0, 1, 2]
    assert _stack_sort([0, 2, 1]) == [0, 1, 2]
    assert _stack_sort([1, 0, 2]) == [0, 1, 2]
    assert _stack_sort([1, 2, 0]) == [1, 0, 2]
    assert _stack_sort([2, 0, 1]) == [0, 1, 2]
    assert _stack_sort([2, 1, 0]) == [0, 1, 2]
    assert _stack_sort([0, 1, 2, 3]) == [0, 1, 2, 3]
    assert _stack_sort([0, 1, 3, 2]) == [0, 1, 2, 3]
    assert _stack_sort([0, 2, 1, 3]) == [0, 1, 2, 3]
    assert _stack_sort([0, 2, 3, 1]) == [0, 2, 1, 3]
    assert _stack_sort([0, 3, 1, 2]) == [0, 1, 2, 3]
    assert _stack_sort([0, 3, 2, 1]) == [0, 1, 2, 3]
    assert _stack_sort([1, 0, 2, 3]) == [0, 1, 2, 3]
    assert _stack_sort([1, 0, 3, 2]) == [0, 1, 2, 3]
    assert _stack_sort([1, 2, 0, 3]) == [1, 0, 2, 3]
    assert _stack_sort([1, 2, 3, 0]) == [1, 2, 0, 3]
    assert _stack_sort([1, 3, 0, 2]) == [1, 0, 2, 3]
    assert _stack_sort([1, 3, 2, 0]) == [1, 0, 2, 3]
    assert _stack_sort([2, 0, 1, 3]) == [0, 1, 2, 3]
    assert _stack_sort([2, 0, 3, 1]) == [0, 2, 1, 3]
    assert _stack_sort([2, 1, 0, 3]) == [0, 1, 2, 3]
    assert _stack_sort([2, 1, 3, 0]) == [1, 2, 0, 3]
    assert _stack_sort([2, 3, 0, 1]) == [2, 0, 1, 3]
    assert _stack_sort([2, 3, 1, 0]) == [2, 0, 1, 3]
    assert _stack_sort([3, 0, 1, 2]) == [0, 1, 2, 3]
    assert _stack_sort([3, 0, 2, 1]) == [0, 1, 2, 3]
    assert _stack_sort([3, 1, 0, 2]) == [0, 1, 2, 3]
    assert _stack_sort([3, 1, 2, 0]) == [1, 0, 2, 3]
    assert _stack_sort([3, 2, 0, 1]) == [0, 1, 2, 3]
    assert _stack_sort([3, 2, 1, 0]) == [0, 1, 2, 3]
    patt231 = Perm((1, 2, 0))
    for _ in range(100):
        p = Perm.random(randint(5, 20))
        if p.avoids(patt231):
            assert list(_stack_sort(list(p))) == sorted(range(len(p)))
        else:
            assert not list(_stack_sort(list(p))) == sorted(range(len(p)))


def test_stack_sortable():
    assert stack_sortable(Perm(()))
    assert stack_sortable(Perm((0,)))
    assert stack_sortable(Perm((0, 1)))
    assert stack_sortable(Perm((1, 0)))
    assert stack_sortable(Perm((0, 1, 2)))
    assert stack_sortable(Perm((0, 2, 1)))
    assert stack_sortable(Perm((1, 0, 2)))
    assert not stack_sortable(Perm((1, 2, 0)))
    assert stack_sortable(Perm((2, 0, 1)))
    assert stack_sortable(Perm((2, 1, 0)))
    assert stack_sortable(Perm((2, 1, 0, 4, 3)))
    assert not stack_sortable(Perm((1, 4, 3, 0, 8, 5, 2, 9, 7, 6)))
    assert not stack_sortable(Perm((1, 3, 0, 2)))
    assert stack_sortable(Perm((3, 0, 1, 2)))
    assert not stack_sortable(Perm((6, 2, 4, 3, 7, 1, 5, 0)))
    assert not stack_sortable(Perm((4, 8, 5, 0, 6, 1, 7, 3, 2)))
    assert stack_sortable(Perm((0, 1, 3, 2)))
    assert stack_sortable(Perm((3, 0, 2, 1)))
    assert not stack_sortable(Perm((1, 0, 7, 2, 5, 4, 3, 8, 9, 6)))
    assert not stack_sortable(Perm((4, 1, 5, 2, 6, 0, 3, 8, 7, 9)))
    assert stack_sortable(Perm((0, 12, 11, 10, 9, 1, 5, 4, 2, 3, 6, 7, 8, 14, 13)))


def test__bubble_sort():
    assert _bubble_sort([]) == []
    assert _bubble_sort([0]) == [0]
    assert _bubble_sort([0, 1]) == [0, 1]
    assert _bubble_sort([1, 0]) == [0, 1]
    assert _bubble_sort([0, 1, 2]) == [0, 1, 2]
    assert _bubble_sort([0, 2, 1]) == [0, 1, 2]
    assert _bubble_sort([1, 0, 2]) == [0, 1, 2]
    assert _bubble_sort([1, 2, 0]) == [1, 0, 2]
    assert _bubble_sort([2, 0, 1]) == [0, 1, 2]
    assert _bubble_sort([2, 1, 0]) == [1, 0, 2]
    assert _bubble_sort([0, 3, 2, 4, 1]) == [0, 2, 3, 1, 4]
    assert _bubble_sort([1, 4, 2, 5, 0, 3]) == [1, 2, 4, 0, 3, 5]
    assert _bubble_sort([3, 6, 7, 2, 5, 4, 1, 0]) == [3, 6, 2, 5, 4, 1, 0, 7]
    assert _bubble_sort([2, 6, 7, 8, 5, 1, 9, 4, 3, 0]) == [
        2,
        6,
        7,
        5,
        1,
        8,
        4,
        3,
        0,
        9,
    ]
    assert _bubble_sort([5, 4, 7, 9, 1, 6, 3, 2, 8, 0]) == [
        4,
        5,
        7,
        1,
        6,
        3,
        2,
        8,
        0,
        9,
    ]


def test_bubble_sortable():
    enumeration = [1]
    enumeration.extend([2 ** i for i in range(10)])
    assert bubble_sortable(Perm(()))
    assert bubble_sortable(Perm((0,)))
    assert bubble_sortable(Perm((0, 1)))
    assert bubble_sortable(Perm((1, 0)))
    assert bubble_sortable(Perm((0, 1, 2)))
    assert bubble_sortable(Perm((0, 2, 1)))
    assert bubble_sortable(Perm((1, 0, 2)))
    assert not bubble_sortable(Perm((1, 2, 0)))
    assert bubble_sortable(Perm((2, 0, 1)))
    assert not bubble_sortable(Perm((2, 1, 0)))
    assert bubble_sortable(Perm((2, 0, 1, 3)))
    assert not bubble_sortable(Perm((1, 4, 3, 2, 0)))
    assert not bubble_sortable(Perm((0, 1, 5, 4, 2, 3)))
    assert not bubble_sortable(Perm((1, 3, 4, 0, 6, 2, 5, 7)))
    assert not bubble_sortable(Perm((1, 5, 6, 7, 0, 3, 2, 4)))
    assert not bubble_sortable(Perm((2, 3, 1, 6, 4, 7, 5, 0)))
    assert not bubble_sortable(Perm((2, 5, 1, 4, 7, 0, 6, 3)))
    assert not bubble_sortable(Perm((3, 0, 4, 5, 7, 6, 1, 2)))
    assert not bubble_sortable(Perm((4, 0, 7, 5, 3, 6, 2, 1)))
    assert not bubble_sortable(Perm((6, 5, 2, 0, 3, 7, 1, 8, 4)))
    for i in range(4, 7):
        assert sum(1 for p in Perm.of_length(i) if bubble_sortable(p)) == enumeration[i]
    patts = (Perm((1, 2, 0)), Perm((2, 1, 0)))
    for i in range(20):
        p = Perm.random(randint(5, 10))
        if p.avoids(*patts):
            assert bubble_sortable(p)
        else:
            assert not bubble_sortable(p)


def test__quick_sort():
    assert _quick_sort([]) == []
    assert _quick_sort([0]) == [0]
    assert _quick_sort([0, 1]) == [0, 1]
    assert _quick_sort([1, 0]) == [0, 1]
    assert _quick_sort([0, 1, 2]) == [0, 1, 2]
    assert _quick_sort([0, 2, 1]) == [0, 1, 2]
    assert _quick_sort([1, 0, 2]) == [0, 1, 2]
    assert _quick_sort([1, 2, 0]) == [0, 1, 2]
    assert _quick_sort([2, 0, 1]) == [0, 1, 2]
    assert _quick_sort([2, 1, 0]) == [1, 0, 2]
    assert _quick_sort([0, 1, 2, 3]) == [0, 1, 2, 3]
    assert _quick_sort([0, 1, 2, 4, 3]) == [0, 1, 2, 3, 4]
    assert _quick_sort([2, 3, 5, 4, 0, 1]) == [0, 1, 2, 3, 5, 4]
    assert _quick_sort([3, 4, 1, 2, 0, 5]) == [1, 2, 0, 3, 4, 5]
    assert _quick_sort([3, 1, 2, 0, 6, 4, 5]) == [1, 2, 0, 3, 6, 4, 5]
    assert _quick_sort([5, 6, 1, 2, 0, 3, 4]) == [1, 2, 0, 3, 4, 5, 6]
    assert _quick_sort([6, 5, 3, 2, 1, 0, 4]) == [5, 3, 2, 1, 0, 4, 6]
    assert _quick_sort([0, 3, 8, 5, 2, 1, 7, 6, 4]) == [0, 2, 1, 3, 8, 5, 7, 6, 4]
    assert _quick_sort([4, 6, 0, 5, 7, 1, 3, 2, 8]) == [0, 1, 3, 2, 4, 6, 5, 7, 8]
    assert _quick_sort([4, 0, 3, 7, 1, 8, 2, 9, 5, 6]) == [0, 3, 1, 2, 4, 7, 8, 9, 5, 6]


def test_quick_sortable():
    enumeration = [1, 1, 2, 5, 12, 28, 65, 151, 351, 816]
    for i in range(4, 7):
        assert sum(1 for p in Perm.of_length(i) if quick_sortable(p)) == enumeration[i]
    assert quick_sortable(Perm(()))
    assert quick_sortable(Perm((0,)))
    assert quick_sortable(Perm((0, 1)))
    assert quick_sortable(Perm((1, 0)))
    assert quick_sortable(Perm((0, 1, 2)))
    assert quick_sortable(Perm((0, 2, 1)))
    assert quick_sortable(Perm((1, 0, 2)))
    assert quick_sortable(Perm((1, 2, 0)))
    assert quick_sortable(Perm((2, 0, 1)))
    assert not quick_sortable(Perm((2, 1, 0)))
    assert not quick_sortable(Perm((2, 1, 0, 3)))
    assert not quick_sortable(Perm((2, 3, 1, 0)))
    assert not quick_sortable(Perm((3, 4, 2, 1, 0)))
    assert not quick_sortable(Perm((4, 3, 2, 0, 1)))
    assert quick_sortable(Perm((0, 2, 1, 3, 4, 5)))
    assert quick_sortable(Perm((0, 2, 3, 4, 5, 1)))
    assert not quick_sortable(Perm((5, 0, 1, 4, 3, 2, 6)))
    assert not quick_sortable(Perm((3, 6, 5, 4, 7, 2, 0, 1)))
    assert not quick_sortable(Perm((6, 9, 7, 2, 3, 0, 5, 4, 1, 8)))
    assert not quick_sortable(Perm((7, 5, 0, 4, 3, 1, 9, 2, 6, 8)))


def test_bkv_sortable():
    patts_to_enumeration = {
        (Perm((0, 1, 2)), Perm((0, 2, 1))): [1, 1, 2, 5, 14, 42, 132, 429],
        (Perm((1, 2, 0)), Perm((0, 2, 1))): [1, 1, 2, 6, 22, 90, 394, 1806],
        (): [1, 1, 2, 5, 14, 42, 132, 429],
        (Perm((2, 1, 0)),): [1, 1, 2, 4, 8, 16, 32, 64],
        (Perm((2, 1, 0, 3)),): [1, 1, 2, 5, 13, 34, 89, 233],
        (Perm((3, 1, 0, 2)),): [1, 1, 2, 5, 13, 34, 89, 233],
        (Perm((3, 2, 0, 1)),): [1, 1, 2, 5, 13, 34, 89, 233],
        (Perm((3, 2, 1, 0)),): [1, 1, 2, 5, 13, 34, 89, 233],
        (Perm((2, 1, 0, 3, 4)),): [1, 1, 2, 5, 14, 41, 121, 355],
        (Perm((4, 1, 0, 2, 3)),): [1, 1, 2, 5, 14, 41, 121, 355],
        (Perm((4, 3, 0, 1, 2)),): [1, 1, 2, 5, 14, 41, 121, 356],
        (Perm((2, 1, 0, 4, 3)),): [1, 1, 2, 5, 14, 41, 122, 365],
        (Perm((3, 1, 0, 2, 4)),): [1, 1, 2, 5, 14, 41, 122, 365],
        (Perm((3, 2, 0, 1, 4)),): [1, 1, 2, 5, 14, 41, 122, 365],
        (Perm((3, 2, 1, 0, 4)),): [1, 1, 2, 5, 14, 41, 122, 365],
        (Perm((4, 1, 0, 3, 2)),): [1, 1, 2, 5, 14, 41, 122, 365],
        (Perm((4, 2, 0, 1, 3)),): [1, 1, 2, 5, 14, 41, 122, 365],
        (Perm((4, 2, 1, 0, 3)),): [1, 1, 2, 5, 14, 41, 122, 365],
        (Perm((4, 3, 0, 2, 1)),): [1, 1, 2, 5, 14, 41, 122, 365],
        (Perm((4, 3, 1, 0, 2)),): [1, 1, 2, 5, 14, 41, 122, 365],
        (Perm((4, 3, 2, 0, 1)),): [1, 1, 2, 5, 14, 41, 122, 365],
        (Perm((4, 3, 2, 1, 0)),): [1, 1, 2, 5, 14, 41, 122, 365],
    }

    all_patts = patts_to_enumeration.keys()
    compares = 7  # can be increased to 8 at most but slow
    res = {k: [0] * compares for k in all_patts}
    for i in range(compares):
        for perm in Perm.of_length(i):
            for k in all_patts:
                if bkv_sortable(perm, k):
                    res[k][i] += 1
    slicer = compares - 8
    if slicer == 0:
        slicer = 8
    assert all(res[pat] == patts_to_enumeration[pat][:slicer] for pat in all_patts)
    assert bkv_sortable(Perm((0, 1)), ())
    assert not bkv_sortable(Perm((7, 6, 9, 2, 0, 3, 8, 1, 5, 4)), (Perm((0, 2, 1)),))
    assert not bkv_sortable(Perm((4, 5, 2, 6, 7, 8, 1, 0, 3)), (Perm((0, 1)),))
    assert bkv_sortable(Perm((2, 0, 1, 3, 4, 5, 6)), (Perm((2, 0, 1, 3)),))
    assert not bkv_sortable(
        Perm((5, 2, 1, 0, 3, 4)), (Perm((2, 1, 0)), Perm((0, 2, 1)))
    )
    assert bkv_sortable(Perm((3, 2, 1, 0, 4)), ())
    assert not bkv_sortable(Perm((4, 1, 2, 3, 0, 5)), (Perm((0, 2, 1)),))
    assert bkv_sortable(Perm((2, 0, 1)), (Perm((0, 1)),))
    assert bkv_sortable(Perm((2, 1, 0)), (Perm((2, 0, 1, 3)),))
    assert bkv_sortable(Perm((5, 3, 1, 0, 4, 2)), (Perm((2, 1, 0)), Perm((0, 2, 1))))
    assert bkv_sortable(Perm((0,)), ())
    assert bkv_sortable(Perm((2, 1, 3, 0)), (Perm((0, 2, 1)),))
    assert bkv_sortable(Perm((0,)), (Perm((0, 1)),))
    assert not bkv_sortable(Perm((6, 4, 0, 5, 8, 3, 2, 7, 1)), (Perm((2, 0, 1, 3)),))
    assert bkv_sortable(Perm(()), (Perm((2, 1, 0)), Perm((0, 2, 1))))
    assert not bkv_sortable(Perm((4, 1, 3, 2, 0)), ())
    assert not bkv_sortable(Perm((1, 2, 0, 3)), (Perm((0, 2, 1)),))
    assert bkv_sortable(Perm(()), (Perm((0, 1)),))
    assert not bkv_sortable(Perm((1, 5, 4, 2, 0, 3)), (Perm((2, 0, 1, 3)),))
    assert bkv_sortable(Perm(()), (Perm((2, 1, 0)), Perm((0, 2, 1))))


def test_west_2_stack_sortable():
    eq_patt_avoidance = (Perm((1, 2, 3, 0)), MeshPatt(Perm((2, 1, 3, 0)), {(1, 4)}))
    for _ in range(100):
        if randint(0, 10) < 1:
            p = Perm.random(randint(0, 3))
        else:
            p = Perm.random(randint(4, 10))
        if p.avoids(*eq_patt_avoidance):
            assert west_2_stack_sortable(p)
        else:
            assert not west_2_stack_sortable(p)
    for n in (5, 6, 7):
        assert sum(
            1 for p in Perm.of_length(n) if west_2_stack_sortable(p)
        ) == 2 * factorial(3 * n) / (factorial(n + 1) * factorial(2 * n + 1))


def test_west_3_stack_sortable():
    assert west_3_stack_sortable(Perm(()))
    assert west_3_stack_sortable(Perm((0,)))
    assert west_3_stack_sortable(Perm((0, 1)))
    assert west_3_stack_sortable(Perm((1, 0)))
    assert west_3_stack_sortable(Perm((0, 1, 2)))
    assert west_3_stack_sortable(Perm((0, 2, 1)))
    assert west_3_stack_sortable(Perm((1, 0, 2)))
    assert west_3_stack_sortable(Perm((1, 2, 0)))
    assert west_3_stack_sortable(Perm((2, 0, 1)))
    assert west_3_stack_sortable(Perm((2, 1, 0)))
    assert west_3_stack_sortable(Perm((0, 1, 2, 3)))
    assert west_3_stack_sortable(Perm((0, 1, 3, 2)))
    assert west_3_stack_sortable(Perm((2, 0, 5, 4, 1, 3)))
    assert not west_3_stack_sortable(Perm((3, 2, 0, 4, 5, 1)))
    assert west_3_stack_sortable(Perm((4, 6, 3, 2, 1, 0, 5)))
    assert west_3_stack_sortable(Perm((6, 2, 3, 0, 5, 1, 4)))
    assert west_3_stack_sortable(Perm((0, 3, 6, 2, 5, 1, 7, 4)))
    assert west_3_stack_sortable(Perm((0, 3, 7, 8, 1, 6, 4, 2, 5)))
    assert not west_3_stack_sortable(Perm((0, 2, 3, 4, 6, 5, 1)))
    assert not west_3_stack_sortable(Perm((2, 3, 0, 4, 6, 1, 5)))
    assert not west_3_stack_sortable(Perm((5, 4, 8, 0, 3, 6, 7, 2, 1)))
    assert not west_3_stack_sortable(Perm((0, 3, 7, 5, 1, 4, 8, 6, 2, 9)))
    assert not west_3_stack_sortable(Perm((5, 1, 11, 10, 9, 8, 6, 0, 4, 3, 7, 2)))
    assert not west_3_stack_sortable(
        Perm((4, 6, 9, 10, 13, 2, 0, 5, 8, 12, 3, 1, 11, 7))
    )
    assert 3494 == sum(1 for p in Perm.of_length(7) if west_3_stack_sortable(p))


def test_smooth():
    assert smooth(Perm(()))
    assert smooth(Perm((0,)))
    assert smooth(Perm((0, 1)))
    assert smooth(Perm((1, 0)))
    assert smooth(Perm((0, 1, 2)))
    assert smooth(Perm((0, 2, 1)))
    assert smooth(Perm((1, 0, 2)))
    assert smooth(Perm((1, 2, 0)))
    assert smooth(Perm((2, 0, 1)))
    assert smooth(Perm((2, 1, 0)))
    assert smooth(Perm((0, 1, 2, 3)))
    assert smooth(Perm((0, 1, 3, 2)))
    assert not smooth(Perm((0, 3, 5, 1, 2, 4)))
    assert not smooth(Perm((3, 4, 1, 0, 5, 2)))
    assert not smooth(Perm((0, 3, 2, 4, 5, 6, 1)))
    assert not smooth(Perm((5, 3, 1, 4, 2, 6, 0)))
    assert smooth(Perm((3, 7, 2, 0, 4, 5, 6, 1)))
    assert not smooth(Perm((4, 7, 1, 3, 8, 5, 0, 6, 2)))
    assert 1552 == sum(1 for p in Perm.of_length(7) if smooth(p))


def test_forest_like():
    assert forest_like(Perm(()))
    assert forest_like(Perm((0,)))
    assert forest_like(Perm((0, 1)))
    assert forest_like(Perm((1, 0)))
    assert forest_like(Perm((0, 1, 2)))
    assert forest_like(Perm((0, 2, 1)))
    assert forest_like(Perm((1, 0, 2)))
    assert forest_like(Perm((1, 2, 0)))
    assert forest_like(Perm((2, 0, 1)))
    assert forest_like(Perm((2, 1, 0)))
    assert forest_like(Perm((0, 1, 2, 3)))
    assert forest_like(Perm((0, 1, 3, 2)))
    assert not forest_like(Perm((3, 4, 1, 0, 5, 2)))
    assert forest_like(Perm((5, 4, 0, 1, 3, 2)))
    assert forest_like(Perm((6, 0, 1, 5, 2, 3, 4)))
    assert forest_like(Perm((6, 1, 2, 5, 3, 4, 0)))
    assert forest_like(Perm((7, 2, 3, 4, 6, 5, 1, 0)))
    assert not forest_like(Perm((0, 1, 6, 7, 4, 2, 3, 8, 5)))
    assert 1661 == sum(1 for p in Perm.of_length(7) if forest_like(p))


def test_baxter():
    assert baxter(Perm(()))
    assert baxter(Perm((0,)))
    assert baxter(Perm((0, 1)))
    assert baxter(Perm((1, 0)))
    assert baxter(Perm((0, 1, 2)))
    assert baxter(Perm((0, 2, 1)))
    assert baxter(Perm((1, 0, 2)))
    assert baxter(Perm((1, 2, 0)))
    assert baxter(Perm((2, 0, 1)))
    assert baxter(Perm((2, 1, 0)))
    assert baxter(Perm((0, 1, 2, 3)))
    assert baxter(Perm((0, 1, 3, 2)))
    assert not baxter(Perm((1, 3, 0, 5, 4, 2)))
    assert not baxter(Perm((5, 3, 1, 0, 4, 2)))
    assert baxter(Perm((5, 4, 3, 0, 1, 2, 6)))
    assert baxter(Perm((6, 0, 5, 3, 4, 1, 2)))
    assert baxter(Perm((5, 3, 4, 2, 0, 1, 6, 7)))
    assert baxter(Perm((1, 2, 4, 6, 5, 3, 0, 8, 7)))
    assert 2074 == sum(1 for p in Perm.of_length(7) if baxter(p))


def test_simsun():
    assert simsun(Perm(()))
    assert simsun(Perm((0,)))
    assert simsun(Perm((0, 1)))
    assert simsun(Perm((1, 0)))
    assert simsun(Perm((0, 1, 2)))
    assert simsun(Perm((0, 2, 1)))
    assert simsun(Perm((1, 0, 2)))
    assert simsun(Perm((1, 2, 0)))
    assert simsun(Perm((2, 0, 1)))
    assert not simsun(Perm((2, 1, 0)))
    assert simsun(Perm((0, 1, 2, 3)))
    assert simsun(Perm((0, 1, 3, 2)))
    assert not simsun(Perm((0, 5, 1, 4, 2, 3)))
    assert not simsun(Perm((1, 3, 5, 4, 2, 0)))
    assert simsun(Perm((1, 5, 0, 2, 3, 6, 4)))
    assert not simsun(Perm((3, 1, 2, 6, 5, 0, 4)))
    assert not simsun(Perm((4, 7, 3, 0, 6, 5, 2, 1)))
    assert not simsun(Perm((4, 6, 7, 0, 3, 5, 2, 8, 1)))
    assert 429 == sum(1 for p in Perm.of_length(7) if simsun(p))


def test_dihedral():
    assert not dihedral(Perm(()))
    assert not dihedral(Perm((0,)))
    assert not dihedral(Perm((0, 1)))
    assert not dihedral(Perm((1, 0)))
    assert dihedral(Perm((0, 1, 2)))
    assert dihedral(Perm((0, 2, 1)))
    assert dihedral(Perm((1, 0, 2)))
    assert dihedral(Perm((1, 2, 0)))
    assert dihedral(Perm((2, 0, 1)))
    assert dihedral(Perm((2, 1, 0)))
    assert dihedral(Perm((0, 1, 2, 3)))
    assert not dihedral(Perm((0, 1, 3, 2)))
    assert not dihedral(Perm((3, 4, 1, 0, 2, 5)))
    assert not dihedral(Perm((5, 4, 2, 1, 0, 3)))
    assert not dihedral(Perm((2, 3, 6, 1, 0, 5, 4)))
    assert not dihedral(Perm((6, 4, 2, 5, 1, 3, 0)))
    assert not dihedral(Perm((1, 6, 4, 2, 0, 5, 3, 7)))
    assert not dihedral(Perm((2, 0, 3, 8, 1, 5, 7, 4, 6)))
    assert not dihedral(Perm((1, 7, 4, 3, 8, 9, 6, 2, 5, 0)))
    assert not dihedral(Perm((5, 6, 2, 10, 1, 0, 9, 3, 7, 4, 8)))
    assert all(
        2 * n == sum(1 for perm in Perm.of_length(n) if dihedral(perm))
        for n in range(5, 8)
    )


def test_in_alternating_group():
    assert in_alternating_group(Perm(()))
    assert in_alternating_group(Perm((0,)))
    assert not in_alternating_group(Perm((0, 1)))
    assert not in_alternating_group(Perm((1, 0)))
    assert in_alternating_group(Perm((0, 1, 2)))
    assert not in_alternating_group(Perm((0, 2, 1)))
    assert not in_alternating_group(Perm((1, 0, 2)))
    assert in_alternating_group(Perm((1, 2, 0)))
    assert in_alternating_group(Perm((2, 0, 1)))
    assert not in_alternating_group(Perm((2, 1, 0)))
    assert in_alternating_group(Perm((0, 1, 2, 3)))
    assert not in_alternating_group(Perm((0, 1, 3, 2)))
    assert not in_alternating_group(Perm((0, 4, 1, 5, 2, 3)))
    assert in_alternating_group(Perm((5, 4, 0, 2, 1, 3)))
    assert not in_alternating_group(Perm((0, 4, 3, 5, 2, 6, 1)))
    assert not in_alternating_group(Perm((1, 3, 0, 2, 5, 6, 4)))
    assert in_alternating_group(Perm((5, 7, 2, 1, 4, 0, 3, 6)))
    assert not in_alternating_group(Perm((3, 0, 6, 1, 2, 8, 5, 4, 7)))
    assert in_alternating_group(Perm((5, 1, 6, 4, 9, 8, 2, 0, 7, 3)))
    assert not in_alternating_group(Perm((7, 0, 2, 4, 8, 5, 1, 9, 6, 3, 10)))
    assert 2520 == sum(1 for p in Perm.of_length(7) if in_alternating_group(p))


def test__perm_to_yt():
    z = {i: set() for i in range(8)}
    for i in range(8):
        for p in Perm.of_length(i):
            z[i].add(tuple(map(tuple, _perm_to_yt(p))))
    expected = [1, 1, 2, 4, 10, 26, 76, 232]
    for i in range(8):
        assert len(z[i]) == expected[i]
    assert _perm_to_yt(Perm((1, 0))) == [[0], [1]]
    assert _perm_to_yt(Perm((1, 0, 3, 5, 2, 4))) == [[0, 2, 4], [1, 3, 5]]
    assert _perm_to_yt(Perm((1, 0, 2, 3))) == [[0, 2, 3], [1]]
    assert _perm_to_yt(Perm((3, 1, 4, 0, 2))) == [[0, 2], [1, 4], [3]]
    assert _perm_to_yt(Perm((5, 4, 1, 0, 3, 2))) == [[0, 2], [1, 3], [4], [5]]
    assert _perm_to_yt(Perm((0, 1))) == [[0, 1]]
    assert _perm_to_yt(Perm((0,))) == [[0]]
    assert _perm_to_yt(Perm((2, 3, 0, 1))) == [[0, 1], [2, 3]]
    assert _perm_to_yt(Perm((1, 2, 3, 0))) == [[0, 2, 3], [1]]
    assert _perm_to_yt(Perm((4, 0, 1, 2, 5, 3))) == [[0, 1, 2, 3], [4, 5]]


def test_yt_perm_avoids_22():
    assert yt_perm_avoids_22(Perm())
    assert yt_perm_avoids_22(Perm((0,)))
    assert yt_perm_avoids_22(Perm((0, 1)))
    assert yt_perm_avoids_22(Perm((1, 0)))
    assert yt_perm_avoids_22(Perm((0, 1, 2)))
    assert yt_perm_avoids_22(Perm((0, 2, 1)))
    assert yt_perm_avoids_22(Perm((1, 0, 2)))
    assert yt_perm_avoids_22(Perm((1, 2, 0)))
    assert yt_perm_avoids_22(Perm((2, 0, 1)))
    assert yt_perm_avoids_22(Perm((2, 1, 0)))
    assert yt_perm_avoids_22(Perm((0, 1, 2, 3)))
    assert yt_perm_avoids_22(Perm((0, 1, 3, 2)))
    assert yt_perm_avoids_22(Perm((2, 5, 3, 1, 0, 4)))
    assert yt_perm_avoids_22(Perm((4, 3, 1, 2, 0, 5)))
    assert not yt_perm_avoids_22(Perm((3, 0, 4, 1, 5, 6, 2)))
    assert not yt_perm_avoids_22(Perm((5, 6, 0, 2, 4, 3, 1)))
    assert not yt_perm_avoids_22(Perm((7, 4, 3, 5, 0, 2, 6, 1)))
    assert yt_perm_avoids_22(Perm((3, 4, 5, 6, 2, 7, 1, 0, 8)))
    assert not yt_perm_avoids_22(Perm((7, 5, 2, 0, 4, 8, 6, 9, 1, 3)))
    assert not yt_perm_avoids_22(Perm((3, 5, 4, 6, 0, 7, 2, 10, 1, 8, 9)))
    assert not yt_perm_avoids_22(Perm((1, 2, 0, 4, 3, 5)))
    assert not yt_perm_avoids_22(Perm((5, 2, 0, 3, 1, 4)))
    assert not yt_perm_avoids_22(Perm((1, 0, 6, 4, 3, 5, 2)))
    assert not yt_perm_avoids_22(Perm((5, 0, 1, 3, 6, 2, 4)))
    assert not yt_perm_avoids_22(Perm((5, 1, 2, 7, 4, 3, 0, 6)))
    assert not yt_perm_avoids_22(Perm((5, 6, 0, 8, 3, 7, 4, 1, 2)))
    assert not yt_perm_avoids_22(Perm((2, 5, 7, 6, 1, 8, 9, 3, 0, 4)))
    assert not yt_perm_avoids_22(Perm((7, 8, 4, 0, 1, 5, 2, 10, 3, 9, 6)))
    assert 924 == sum(1 for p in Perm.of_length(7) if yt_perm_avoids_22(p))


def test_yt_perm_avoids_32():
    assert yt_perm_avoids_32(Perm())
    assert yt_perm_avoids_32(Perm((0,)))
    assert yt_perm_avoids_32(Perm((0, 1)))
    assert yt_perm_avoids_32(Perm((1, 0)))
    assert yt_perm_avoids_32(Perm((0, 1, 2)))
    assert yt_perm_avoids_32(Perm((0, 2, 1)))
    assert yt_perm_avoids_32(Perm((1, 0, 2)))
    assert yt_perm_avoids_32(Perm((1, 2, 0)))
    assert yt_perm_avoids_32(Perm((2, 0, 1)))
    assert yt_perm_avoids_32(Perm((2, 1, 0)))
    assert yt_perm_avoids_32(Perm((0, 1, 2, 3)))
    assert yt_perm_avoids_32(Perm((0, 1, 3, 2)))
    assert yt_perm_avoids_32(Perm((2, 3, 1, 4, 0, 5)))
    assert not yt_perm_avoids_32(Perm((2, 3, 1, 5, 4, 0)))
    assert not yt_perm_avoids_32(Perm((2, 0, 3, 1, 4, 5, 6)))
    assert not yt_perm_avoids_32(Perm((5, 1, 2, 0, 4, 3, 6)))
    assert not yt_perm_avoids_32(Perm((5, 0, 4, 2, 1, 6, 3, 7)))
    assert not yt_perm_avoids_32(Perm((0, 6, 3, 1, 8, 7, 4, 5, 2)))
    assert not yt_perm_avoids_32(Perm((6, 4, 9, 0, 3, 1, 8, 7, 5, 2)))
    assert not yt_perm_avoids_32(Perm((10, 8, 1, 4, 0, 5, 2, 7, 9, 3, 6)))
    assert yt_perm_avoids_32(Perm((7, 3, 2, 4, 5, 1, 0, 6)))
    assert yt_perm_avoids_32(Perm((5, 4, 6, 1, 3, 2, 0)))
    assert yt_perm_avoids_32(Perm((4, 6, 2, 0, 5, 3, 1)))
    assert 1316 == sum(1 for p in Perm.of_length(7) if yt_perm_avoids_32(p))
    assert yt_perm_avoids_32(Perm((0, 14, 1, 13, 11, 8, 3, 5, 7, 9, 6, 10, 12, 4, 2)))
    assert yt_perm_avoids_32(Perm((7, 5, 2, 4, 6, 8, 9, 10, 3, 11, 12, 1, 0, 13, 14)))
    assert yt_perm_avoids_32(Perm((13, 9, 8, 0, 5, 1, 3, 4, 2, 6, 7, 10, 11, 12, 14)))
    assert yt_perm_avoids_32(Perm((14, 10, 13, 12, 8, 7, 11, 5, 9, 6, 3, 4, 1, 0, 2)))
    assert yt_perm_avoids_32(Perm((13, 0, 1, 10, 8, 7, 9, 11, 6, 5, 4, 12, 3, 14, 2)))
    assert yt_perm_avoids_32(Perm((0, 14, 1, 3, 12, 4, 5, 6, 10, 11, 9, 8, 7, 13, 2)))
    assert yt_perm_avoids_32(Perm((12, 9, 14, 13, 11, 6, 10, 8, 5, 2, 1, 7, 4, 0, 3)))
    assert yt_perm_avoids_32(Perm((7, 14, 13, 6, 4, 12, 11, 3, 10, 9, 8, 1, 0, 5, 2)))


def test_av_231_and_mesh():
    assert av_231_and_mesh(Perm(()))
    assert av_231_and_mesh(Perm((0,)))
    assert av_231_and_mesh(Perm((0, 1)))
    assert av_231_and_mesh(Perm((1, 0)))
    assert av_231_and_mesh(Perm((0, 1, 2)))
    assert av_231_and_mesh(Perm((0, 2, 1)))
    assert av_231_and_mesh(Perm((1, 0, 2)))
    assert not av_231_and_mesh(Perm((1, 2, 0)))
    assert av_231_and_mesh(Perm((2, 0, 1)))
    assert av_231_and_mesh(Perm((2, 1, 0)))
    assert av_231_and_mesh(Perm((0, 1, 2, 3)))
    assert av_231_and_mesh(Perm((0, 1, 3, 2)))
    assert av_231_and_mesh(Perm((0, 5, 4, 3, 2, 1)))
    assert av_231_and_mesh(Perm((5, 2, 1, 0, 4, 3)))
    assert not av_231_and_mesh(Perm((3, 4, 6, 2, 5, 0, 1)))
    assert not av_231_and_mesh(Perm((5, 0, 1, 6, 3, 4, 2)))
    assert not av_231_and_mesh(Perm((2, 5, 0, 7, 3, 4, 6, 1)))
    assert not av_231_and_mesh(Perm((5, 7, 4, 8, 1, 6, 0, 2, 3)))
    assert not av_231_and_mesh(Perm((2, 5, 9, 1, 7, 3, 8, 0, 4, 6)))
    assert not av_231_and_mesh(Perm((5, 6, 0, 3, 2, 1, 10, 8, 7, 4, 9)))
    assert av_231_and_mesh(Perm((4, 1, 0, 3, 2, 5, 6)))
    assert not av_231_and_mesh(Perm((2, 6, 4, 0, 1, 7, 3, 5)))
    assert av_231_and_mesh(Perm((9, 8, 6, 5, 1, 0, 4, 3, 2, 7)))
    assert av_231_and_mesh(Perm((0, 3, 2, 1, 9, 8, 7, 6, 5, 4)))
    assert av_231_and_mesh(Perm((9, 1, 0, 2, 4, 3, 7, 5, 6, 8)))
    assert av_231_and_mesh(Perm((8, 5, 1, 0, 2, 4, 3, 7, 6, 9)))
    assert av_231_and_mesh(Perm((9, 1, 0, 8, 2, 5, 4, 3, 6, 7)))
    assert av_231_and_mesh(Perm((9, 8, 7, 4, 0, 3, 1, 2, 5, 6)))
    assert av_231_and_mesh(Perm((6, 1, 0, 2, 5, 3, 4, 9, 7, 8)))
    assert av_231_and_mesh(Perm((8, 6, 0, 5, 3, 2, 1, 4, 7, 9)))
    assert av_231_and_mesh(Perm((0, 1, 3, 2, 4, 6, 5, 9, 8, 7)))
    assert 417 == sum(1 for p in Perm.of_length(7) if av_231_and_mesh(p))


def test_hard_mesh():
    assert hard_mesh(Perm(()))
    assert hard_mesh(Perm((0,)))
    assert hard_mesh(Perm((0, 1)))
    assert hard_mesh(Perm((1, 0)))
    assert not hard_mesh(Perm((0, 1, 2)))
    assert hard_mesh(Perm((0, 2, 1)))
    assert hard_mesh(Perm((1, 0, 2)))
    assert hard_mesh(Perm((1, 2, 0)))
    assert hard_mesh(Perm((2, 0, 1)))
    assert hard_mesh(Perm((2, 1, 0)))
    assert not hard_mesh(Perm((0, 1, 2, 3)))
    assert not hard_mesh(Perm((0, 1, 3, 2)))
    assert not hard_mesh(Perm((1, 5, 0, 2, 3, 4)))
    assert not hard_mesh(Perm((2, 4, 1, 5, 0, 3)))
    assert hard_mesh(Perm((1, 3, 2, 5, 4, 6, 0)))
    assert not hard_mesh(Perm((4, 3, 1, 6, 2, 0, 5)))
    assert not hard_mesh(Perm((3, 4, 1, 5, 2, 6, 0, 7)))
    assert not hard_mesh(Perm((4, 3, 2, 0, 1, 5, 7, 6, 8)))
    assert not hard_mesh(Perm((7, 0, 5, 4, 6, 3, 2, 9, 1, 8)))
    assert not hard_mesh(Perm((2, 6, 9, 1, 8, 0, 10, 5, 7, 3, 4)))
    assert 692 == sum(1 for p in Perm.of_length(7) if hard_mesh(p))



from collections import deque
from typing import Iterable

from ..patterns import Perm


def dihedral_group(n: int) -> Iterable[Perm]:
    """Generate dihedral group of length n.

    Examples:
        >>> sorted(dihedral_group(4))
        [Perm((0, 1, 2, 3)), Perm((0, 3, 2, 1)), Perm((1, 0, 3, 2)), Perm((1, 2, 3, 0))\
, Perm((2, 1, 0, 3)), Perm((2, 3, 0, 1)), Perm((3, 0, 1, 2)), Perm((3, 2, 1, 0))]
    """
    if n <= 2:
        return
    increasing, decreasing = deque(range(n)), deque(range(n - 1, -1, -1))
    yield Perm(increasing)
    yield Perm(decreasing)
    for _ in range(n - 1):
        increasing.appendleft(increasing.pop())
        decreasing.appendleft(decreasing.pop())
        yield Perm(increasing)
        yield Perm(decreasing)



from typing import Iterable

from permuta.patterns.perm import Iterator, Perm, Set, Tuple


def rotate_90_clockwise_set(perms: Iterable[Perm]) -> Iterator[Perm]:
    """Rotate all perms by 90 degrees."""
    return (perm.rotate() for perm in perms)


def rotate_180_clockwise_set(perms: Iterable[Perm]) -> Iterator[Perm]:
    """Rotate all perms by 180 degrees."""
    return (perm.rotate(2) for perm in perms)


def rotate_270_clockwise_set(perms: Iterable[Perm]) -> Iterator[Perm]:
    """Rotate all perms by 270 degrees."""
    return (perm.rotate(3) for perm in perms)


def inverse_set(perms: Iterable[Perm]) -> Iterator[Perm]:
    """Return the inverse of each permutation a collection of perms."""
    return (perm.inverse() for perm in perms)


def reverse_set(perms: Iterable[Perm]) -> Iterator[Perm]:
    """Return the reverse of each permutation a collection of perms."""
    return (perm.reverse() for perm in perms)


def complement_set(perms: Iterable[Perm]) -> Iterator[Perm]:
    """Return the complement of each permutation in a collection of perms."""
    return (perm.complement() for perm in perms)


def antidiagonal_set(perms: Iterable[Perm]) -> Iterator[Perm]:
    """Return the antidiagonal of each permutation in a collection of perms."""
    return (perm.flip_antidiagonal() for perm in perms)


def all_symmetry_sets(perms: Iterable[Perm]) -> Set[Tuple[Perm, ...]]:
    """Given a collection of perms, return all possible collections formed by applying
    symmetric operations on all perms. Each group has their permutation sorted.
    """
    perms = perms if isinstance(perms, list) else list(perms)
    answer = {tuple(sorted(perms)), tuple(sorted(inverse_set(perms)))}
    for _ in range(3):
        perms = list(rotate_90_clockwise_set(perms))
        answer.add(tuple(sorted(perms)))
        answer.add(tuple(sorted(inverse_set(perms))))
    return answer


def lex_min(perms: Iterable[Perm]) -> Tuple[Perm, ...]:
    """Find the lexicographical minimum of the sets of all symmetries."""
    return min(all_symmetry_sets(perms))



from itertools import tee
from typing import Iterable

from permuta.patterns.perm import Perm


def is_finite(basis: Iterable[Perm]) -> bool:
    """Check if a basis is finite, i.e. it contains decreasing and increasing
    permutations.
    """
    it1, it2 = tee(basis, 2)
    return any(perm.is_decreasing() for perm in it1) and any(
        perm.is_increasing() for perm in it2
    )



import os
import tempfile
import threading
import time
import webbrowser
from typing import ClassVar


class HTMLViewer:
    """A class for opening html text in browser."""

    _THREAD_WAIT_TIME: ClassVar[float] = 5  # seconds

    @staticmethod
    def _remove_file_thread(fname: str) -> None:
        time.sleep(HTMLViewer._THREAD_WAIT_TIME)
        if os.path.exists(fname):
            os.remove(fname)

    @staticmethod
    def _remove_file(fname: str) -> None:
        threading.Thread(target=HTMLViewer._remove_file_thread, args=(fname,)).start()

    @staticmethod
    def open_html(html: str) -> None:
        """Open and render html string in browser."""
        with tempfile.NamedTemporaryFile(
            "r+", suffix=".html", delete=False
        ) as html_file:
            html_file.write(html)
            webbrowser.open_new_tab(f"file://{html_file.name}")
            HTMLViewer._remove_file(html_file.name)

    @staticmethod
    def open_svg(svg: str) -> None:
        """Open and render svg image string in browser."""
        HTMLViewer.open_html(f"<html><body>{svg}</body></html>")



from .bivincularpatt import BivincularPatt, CovincularPatt, VincularPatt
from .meshpatt import MeshPatt
from .patt import Patt
from .perm import Perm

__all__ = [
    "Patt",
    "Perm",
    "MeshPatt",
    "BivincularPatt",
    "VincularPatt",
    "CovincularPatt",
]



from random import randint
from typing import Iterable, Iterator, List, Optional, Tuple

from .meshpatt import MeshPatt
from .patt import Patt
from .perm import Perm


class BivincularPatt(MeshPatt):
    """A bivincular pattern class."""

    @staticmethod
    def _to_shading(
        n: int, adjacent_indices: Iterable[int], adjacent_values: Iterable[int]
    ) -> Iterator[Tuple[int, int]]:
        """Convert adjacent requirements into shading."""
        for idx in adjacent_indices:
            assert 0 <= idx <= n
            yield from ((idx, val) for val in range(n + 1))
        for val in adjacent_values:
            assert 0 <= val <= n
            yield from ((idx, val) for idx in range(n + 1))

    def __init__(
        self,
        perm: Perm,
        adjacent_indices: Iterable[int],
        adjacent_values: Iterable[int],
    ) -> None:
        super().__init__(
            perm,
            BivincularPatt._to_shading(len(perm), adjacent_indices, adjacent_values),
        )

    @classmethod
    def unrank(cls, pattern: Perm, number: int) -> MeshPatt:
        """Not implemented, inherited from MeshPatt."""
        raise NotImplementedError

    @classmethod
    def of_length(
        cls, length: int, patt: Optional[Perm] = None
    ) -> Iterator["MeshPatt"]:
        """Not implemented, inherited from MeshPatt."""
        raise NotImplementedError

    @classmethod
    def random(cls, length: int) -> "BivincularPatt":
        """Return a random Bivincular pattern of a given length."""
        return cls(
            Perm.random(length),
            (
                i
                for i, keep in enumerate(randint(0, 1) for _ in range(length + 1))
                if keep
            ),
            (
                i
                for i, keep in enumerate(randint(0, 1) for _ in range(length + 1))
                if keep
            ),
        )

    def get_adjacent_requirements(self) -> Tuple[List[int], List[int]]:
        """Convert shading into the bivincular requirements. Returned as a tuple of
        adjacent indices and adjacent values, both in order.

        Examples:
            >>> BivincularPatt(Perm((0, 3, 1, 2)), (0, 1, 2),
            ... (0, 2, 4)).get_adjacent_requirements()
            ([0, 1, 2], [0, 2, 4])
        """
        n, adj_idx, adj_val = len(self), set(), set()
        for x, y in self.shading:
            if all((x, i) in self.shading for i in range(n + 1)):
                adj_idx.add(x)
            if all((i, y) in self.shading for i in range(n + 1)):
                adj_val.add(y)
        return sorted(adj_idx), sorted(adj_val)

    def occurrences_in(self, patt: Patt, *args, **kwargs) -> Iterator[Tuple[int, ...]]:
        """Find all indices of self in patt. Each yielded element is a tuple of integer
        indices of the pattern such that

            Classical pattern:
                Occurrence of instance's perm in patt if no elements land
                in shaded region.

            Mesh pattern (including Bivincular):
                Occurrences of instances's perm in the pattern's perm is found, and if
                the sub mesh pattern formed by the occurrence indices is a superset of
                the instance shading, they are included.
        """
        if isinstance(patt, Perm):
            # TODO: Optimize me for Bivincular patterns
            pass
        return super().occurrences_in(patt, args, kwargs)

    def __repr__(self) -> str:
        adj_idx, adj_val = self.get_adjacent_requirements()
        return f"BivincularPatt({repr(self.pattern)}, {adj_idx}, {adj_val})"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, MeshPatt):
            return self.pattern == other.pattern and self.shading == other.shading
        return False

    def __hash__(self) -> int:
        return hash(super())


class VincularPatt(BivincularPatt):
    """A vincular pattern class."""

    def __init__(self, perm: Perm, adjacent_indices: Iterable[int]) -> None:
        super().__init__(perm, adjacent_indices, ())

    @classmethod
    def random(cls, length: int) -> "VincularPatt":
        """Return a random Vincular pattern of a given length."""
        return cls(
            Perm.random(length),
            (
                i
                for i, keep in enumerate(randint(0, 1) for _ in range(length + 1))
                if keep
            ),
        )

    def __repr__(self) -> str:
        adj_idx, _ = self.get_adjacent_requirements()
        return f"VincularPatt({repr(self.pattern)}, {adj_idx})"


class CovincularPatt(BivincularPatt):
    """A covincular pattern class."""

    def __init__(self, perm: Perm, adjacent_values: Iterable[int]) -> None:
        super().__init__(perm, (), adjacent_values)

    @classmethod
    def random(cls, length: int) -> "CovincularPatt":
        """Return a random Covincular pattern of a given length."""
        return cls(
            Perm.random(length),
            (
                i
                for i, keep in enumerate(randint(0, 1) for _ in range(length + 1))
                if keep
            ),
        )

    def __repr__(self) -> str:
        _, adj_val = self.get_adjacent_requirements()
        return f"CovincularPatt({repr(self.pattern)}, {adj_val})"



import argparse
import signal
import sys
import types

from permuta import Av, Basis
from permuta.permutils import InsertionEncodablePerms, PolyPerms, lex_min


def sigint_handler(sig: int, _frame: types.FrameType) -> None:
    """For terminating infinite task."""
    if sig == signal.SIGINT:
        print("\nExiting.")
        sys.exit(0)


def enumerate_class(args: argparse.Namespace) -> None:
    """Enumerate a perm class indefinitely, one length at a time."""
    signal.signal(signal.SIGINT, sigint_handler)
    perm_class = Av.from_string(args.basis)
    print(f"Enumerating {perm_class}. Press Ctrl+C to exit.")
    n = 0
    while True:
        print(perm_class.count(n), end=", ", flush=True)
        n += 1


def has_regular_insertion_encoding(args: argparse.Namespace) -> None:
    """Check if a perm class has a regular insertion encoding."""
    basis = Basis.from_string(args.basis)
    perm_class = Av(basis)
    if InsertionEncodablePerms.is_insertion_encodable_maximum(basis):
        print(f"The class {perm_class} has a regular topmost insertion encoding")
    if InsertionEncodablePerms.is_insertion_encodable_rightmost(basis):
        print(f"The class {perm_class} has a regular rightmost insertion encoding")
    if not InsertionEncodablePerms.is_insertion_encodable(basis):
        print(f"{perm_class} does not have a regular insertion encoding")


def get_lex_min(args: argparse.Namespace) -> None:
    """Prints the 0-based lexicographically minimal representation of the basis."""
    basis = Basis.from_string(args.basis)
    print("_".join(str(perm) for perm in lex_min(basis)))


def has_poly_growth(args: argparse.Namespace) -> None:
    """Prints whether perm class from basis has polynomial growth."""
    basis = Basis.from_string(args.basis)
    poly = PolyPerms.is_polynomial(basis)
    print(f"Av({basis}) is {'' if poly else 'not '}polynomial")


def get_parser() -> argparse.ArgumentParser:
    """Construct and return parser."""
    basis_str: str = (
        "The basis as a string where the permutations are separated any token, "
        "(e.g. '231_4321', '0132:43210')"
    )

    parser: argparse.ArgumentParser = argparse.ArgumentParser(
        description="A set of tools to work with permutations"
    )
    subparsers = parser.add_subparsers(title="subcommands")

    # The count command
    count_parser: argparse.ArgumentParser = subparsers.add_parser(
        "count",
        description="A tool to quickly get the enumeration of permutation classes",
    )
    count_parser.set_defaults(func=enumerate_class)
    count_parser.add_argument(
        "basis", help=basis_str,
    )

    # The insenc command
    insenc_parser: argparse.ArgumentParser = subparsers.add_parser(
        "insenc",
        description="A tool to check if a permutation class has a regular insertion"
        " encoding.",
    )
    insenc_parser.set_defaults(func=has_regular_insertion_encoding)
    insenc_parser.add_argument(
        "basis", help=basis_str,
    )

    # The lexmin command
    lexmin_parser: argparse.ArgumentParser = subparsers.add_parser(
        "lexmin",
        description="A tool that returns the 0-based lexicographically minimal "
        "representation of the basis.",
    )
    lexmin_parser.set_defaults(func=get_lex_min)
    lexmin_parser.add_argument(
        "basis", help=basis_str,
    )

    # The poly command
    poly_parser: argparse.ArgumentParser = subparsers.add_parser(
        "poly",
        description="A tool to check if permutation class has polynomial growth.",
    )
    poly_parser.set_defaults(func=has_poly_growth)
    poly_parser.add_argument("basis", help=basis_str)

    return parser


def main() -> None:
    """Entry point."""
    parser = get_parser()
    args = parser.parse_args()
    if not hasattr(args, "func"):
        parser.error("Invalid command")
    args.func(args)


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod
from typing import FrozenSet, Iterable, Iterator

from permuta import Perm
from permuta.permutils.symmetry import all_symmetry_sets


class EnumerationStrategy(ABC):
    """Abstract class for a strategy to enumerate a permutation classes."""

    def __init__(self, basis: Iterable[Perm]) -> None:
        self._basis = frozenset(basis)

    @property
    def basis(self) -> FrozenSet[Perm]:
        """Getter for basis."""
        return self._basis

    @classmethod
    def reference(cls) -> str:
        """A reference for the strategy."""
        raise NotImplementedError

    @abstractmethod
    def applies(self) -> bool:
        """Return True if the strategy can be used for the basis."""


class EnumerationStrategyWithSymmetry(EnumerationStrategy):
    """Abstract class for a strategy to enumerate a permutation classes.
    Each symmetry of the inputed basis is tested against the strategy.
    """

    def applies(self) -> bool:
        """Check if the strategy applies to any symmetry."""
        syms: Iterator[FrozenSet[Perm]] = map(frozenset, all_symmetry_sets(self._basis))
        return next((True for b in syms if self._applies_to_symmetry(b)), False)

    @abstractmethod
    def _applies_to_symmetry(self, basis: FrozenSet[Perm]) -> bool:
        """Check if the strategy applies to this particular symmetry."""



from permuta.enumeration_strategies.abstract_strategy import EnumerationStrategy
from permuta.permutils.insertion_encodable import InsertionEncodablePerms
from permuta.permutils.symmetry import rotate_90_clockwise_set


class InsertionEncodingStrategy(EnumerationStrategy):
    """Enumeration strategies related to the insertion encoding."""

    def applies(self) -> bool:
        return InsertionEncodablePerms.is_insertion_encodable(
            self.basis
        ) or InsertionEncodablePerms.is_insertion_encodable(
            rotate_90_clockwise_set(self.basis)
        )



import abc
from typing import TYPE_CHECKING, Iterator, Tuple

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from .perm import Perm


class Patt(abc.ABC):
    """A permutation pattern, e.g. classical, bivincular and mesh patterns."""

    def avoided_by(self, *patts: "Patt") -> bool:
        """Check if self is avoided by all the provided patterns."""
        return all(self not in patt for patt in patts)

    def contained_in(self, *patts: "Patt") -> bool:
        """Check if self is a pattern of all the provided patterns."""
        return all(self in patt for patt in patts)

    def count_occurrences_in(self, patt: "Patt") -> int:
        """Count the number of occurrences of self in the pattern."""
        return sum(1 for _ in self.occurrences_in(patt))

    @abc.abstractmethod
    def occurrences_in(
        self, patt: "Patt", *args, **kwargs
    ) -> Iterator[Tuple[int, ...]]:
        """Find all indices of occurrences of self in pattern."""

    @abc.abstractmethod
    def __len__(self) -> int:
        """The length of the pattern."""

    @abc.abstractmethod
    def get_perm(self) -> "Perm":
        """Get the permutation part of the pattern"""

    @abc.abstractmethod
    def __contains__(self, patt: object) -> bool:
        """Does pattern contains another?"""



import multiprocessing
from itertools import islice
from typing import ClassVar, Dict, Iterable, List, NamedTuple, Optional, Union

from ..patterns import MeshPatt, Perm
from ..permutils import is_finite, is_insertion_encodable, is_polynomial
from .basis import Basis, MeshBasis


class AvBase(NamedTuple):
    """A base class for Av to define instance variables without having to use
    __init__ in Av.
    """

    basis: Union[Basis, MeshBasis]
    cache: List[Dict[Perm, Optional[List[int]]]]


class Av(AvBase):
    """A permutation class defined by its minimal basis."""

    _FORBIDDEN_BASIS = Basis(Perm())
    _VALUE_ERROR_MSG = "Basis should be non-empty without the empty perm!"
    _BASIS_ONLY_MSG = "Only supported for Basis!"
    _CLASS_CACHE: ClassVar[Dict[Union[Basis, MeshBasis], "Av"]] = {}
    _CACHE_LOCK = multiprocessing.Lock()

    def __new__(
        cls,
        basis: Union[
            Basis, MeshBasis, Iterable[Perm], Iterable[Union[Perm, MeshPatt]],
        ],
    ) -> "Av":
        if not isinstance(basis, (Basis, MeshBasis)):
            return Av.from_iterable(basis)
        if len(basis) == 0 or basis == Av._FORBIDDEN_BASIS:
            raise ValueError(Av._VALUE_ERROR_MSG)
        instance = Av._CLASS_CACHE.get(basis)
        if instance is None:
            new_instance: "Av" = AvBase.__new__(cls, basis, [{Perm(): [0]}])
            Av._CLASS_CACHE[basis] = new_instance
            return new_instance
        return instance

    @classmethod
    def clear_cache(cls) -> None:
        """Clear the instance cache."""
        cls._CLASS_CACHE = {}

    @classmethod
    def from_string(cls, basis) -> "Av":
        """Create a permutation class from a string. Basis can be either zero or one
        based and seperated by anything. MeshBasis is not supported.
        """
        return cls(Basis.from_string(basis))

    @classmethod
    def from_iterable(
        cls, basis: Union[Iterable[Perm], Iterable[Union[Perm, MeshPatt]]]
    ) -> "Av":
        """Create a permutation class from a basis defined by an iterable of patterns.
        """
        if MeshBasis.is_mesh_basis(basis):
            return cls(MeshBasis(*basis))
        return cls(Basis(*basis))

    def is_finite(self) -> bool:
        """Check if the perm class is finite."""
        if isinstance(self.basis, MeshBasis):
            raise NotImplementedError(Av._BASIS_ONLY_MSG)
        return is_finite(self.basis)

    def is_polynomial(self) -> bool:
        """Check if the perm class has polynomial growth."""
        if isinstance(self.basis, MeshBasis):
            raise NotImplementedError(Av._BASIS_ONLY_MSG)
        return is_polynomial(self.basis)

    def is_insertion_encodable(self) -> bool:
        """Check if the perm class is insertion encodable."""
        if isinstance(self.basis, MeshBasis):
            raise NotImplementedError(Av._BASIS_ONLY_MSG)
        return is_insertion_encodable(self.basis)

    def first(self, count: int) -> Iterable[Perm]:
        """Generate the first `count` permutation in this permutation class given
        that it has that many, if not all are generated.
        """
        yield from islice(self._all(), count)

    def of_length(self, length: int) -> Iterable[Perm]:
        """Generate all perms of a given length that belong to this permutation class.
        """
        return iter(self._get_level(length))

    def up_to_length(self, length: int) -> Iterable[Perm]:
        """Generate all perms up to and including a given length that
        belong to this permutation class.
        """
        for n in range(length + 1):
            yield from self.of_length(n)

    def count(self, length: int) -> int:
        """Return the nubmber of permutations of a given length."""
        return len(self._get_level(length))

    def enumeration(self, length: int) -> List[int]:
        """Return the enumeration of this permutation class up and including a given
        length."""
        return [self.count(i) for i in range(length + 1)]

    def __contains__(self, other: object):
        if isinstance(other, Perm):
            return other in self._get_level(len(other))
        return False

    def is_subclass(self, other: "Av"):
        """Check if a sublcass of another permutation class."""
        return all(p1 not in self for p1 in other.basis)

    def _ensure_level(self, level_number: int) -> None:
        start = max(0, len(self.cache) - 2)
        if isinstance(self.basis, Basis):
            self._ensure_level_classical_pattern_basis(level_number)
        else:
            self._ensure_level_mesh_pattern_basis(level_number)
        for i in range(start, level_number - 1):
            self.cache[i] = {perm: None for perm in self.cache[i]}

    def _ensure_level_classical_pattern_basis(self, level_number: int) -> None:
        # We build new elements from existing ones
        lengths = {len(b) for b in self.basis}
        max_size = max(lengths)
        for nplusone in range(len(self.cache), level_number + 1):
            n = nplusone - 1
            new_level: Dict[Perm, Optional[List[int]]] = dict()
            last_level = self.cache[-1]
            check_length = nplusone in lengths
            smaller_elems = {b for b in self.basis if len(b) == nplusone}

            def valid_insertions(perm):
                # pylint: disable=cell-var-from-loop
                res = None
                for i in range(max(0, n - max_size), n):
                    val = perm[i]
                    subperm = perm.remove(i)
                    spots = self.cache[n - 1][subperm]
                    acceptable = [k for k in spots if k <= val]
                    acceptable.extend(k + 1 for k in spots if k >= val)
                    if res is None:
                        res = frozenset(acceptable)
                    res = res.intersection(acceptable)
                    if not res:
                        break
                return res if res is not None else range(nplusone)

            for perm, lis in last_level.items():
                for value in valid_insertions(perm):
                    new_perm = perm.insert(index=nplusone, new_element=value)
                    if not check_length or new_perm not in smaller_elems:
                        new_level[new_perm] = []
                        assert lis is not None
                        lis.append(value)
            self.cache.append(new_level)

    def _ensure_level_mesh_pattern_basis(self, level_number: int) -> None:
        self.cache.extend(
            {p: None for p in Perm.of_length(i) if p.avoids(*self.basis)}
            for i in range(len(self.cache), level_number + 1)
        )

    def _get_level(self, level_number: int) -> Dict[Perm, Optional[List[int]]]:
        with Av._CACHE_LOCK:
            self._ensure_level(level_number)
        return self.cache[level_number]

    def _all(self) -> Iterable[Perm]:
        length = 0
        while True:
            gen = (p for p in self.of_length(length))
            first: Optional[Perm] = next(gen, None)
            if first is None:
                break
            yield first
            yield from gen
            length += 1

    def __str__(self) -> str:
        return f"Av({','.join(str(p) for p in self.basis)})"

    def __repr__(self) -> str:
        return f"Av({repr(self.basis)})"



import random
from itertools import combinations

from permuta.misc.union_find import UnionFind


def test_union_find_init():
    uf = UnionFind(10)
    assert sum(x == -1 for x in uf._parent) == 10


def test_union_find_1():
    uf = UnionFind(10)
    assert uf.find(1) != uf.find(3)
    assert uf.unite(1, 8)
    assert uf.unite(3, 8)
    assert uf.find(1) == uf.find(3)
    assert uf.size(1) == uf.size(3) == uf.size(8) == 3
    assert all(uf.size(i) == 1 for i in range(10) if i not in (1, 3, 8))


def test_union_find_2():
    uf = UnionFind(100)
    for i in range(1, 100):
        assert uf.unite(0, i)
    for (a, b) in combinations(range(100), 2):
        assert uf.find(a) == uf.find(b)
        assert uf.size(a) == 100
        assert uf.size(b) == 100


def test_union_find_3():
    uf = UnionFind(4)
    assert uf.find(0) == uf.find(0)
    assert not uf.unite(0, 0)
    assert uf.unite(1, 0)
    assert not uf.unite(0, 1)
    assert uf.unite(1, 2)
    assert not uf.unite(1, 2)
    assert not uf.unite(0, 2)
    assert uf.find(0) != uf.find(3)


def test_union_find_4():
    n = 5000
    d = {i: {i} for i in range(n)}
    uf = UnionFind(n)

    for i in range(2 * n):
        a, b = random.randint(0, n - 1), random.randint(0, n - 1)
        if random.randint(0, 1):
            combined = d[a].union(d[b])
            for x in combined:
                d[x] = combined
            uf.unite(a, b)
            assert uf.size(a) == len(combined) == uf.size(b)
        else:
            if a in d[b]:
                assert uf.find(a) == uf.find(b)
            else:
                assert uf.find(a) != uf.find(b)



from .display import HTMLViewer
from .union_find import UnionFind

DIR_EAST = 0
DIR_NORTH = 1
DIR_WEST = 2
DIR_SOUTH = 3
DIR_NONE = -1
DIRS = [DIR_EAST, DIR_NORTH, DIR_WEST, DIR_SOUTH]

__all__ = [
    "HTMLViewer",
    "UnionFind",
    "DIRS",
    "DIR_EAST",
    "DIR_NORTH",
    "DIR_WEST",
    "DIR_SOUTH",
    "DIR_NONE",
]



from .finite import is_finite
from .groups import dihedral_group
from .insertion_encodable import InsertionEncodablePerms
from .polynomial import PolyPerms
from .symmetry import (
    all_symmetry_sets,
    antidiagonal_set,
    complement_set,
    inverse_set,
    lex_min,
    reverse_set,
    rotate_90_clockwise_set,
    rotate_180_clockwise_set,
    rotate_270_clockwise_set,
)

is_insertion_encodable = InsertionEncodablePerms.is_insertion_encodable
is_insertion_encodable_maximum = InsertionEncodablePerms.is_insertion_encodable_maximum
is_insertion_encodable_rightmost = (
    InsertionEncodablePerms.is_insertion_encodable_rightmost
)
is_polynomial = PolyPerms.is_polynomial
is_non_polynomial = PolyPerms.is_non_polynomial

__all__ = [
    "is_insertion_encodable",
    "is_insertion_encodable_maximum",
    "is_insertion_encodable_rightmost",
    "is_polynomial",
    "is_non_polynomial",
    "InsertionEncodablePerms",
    "PolyPerms",
    "is_finite",
    "dihedral_group",
    "all_symmetry_sets",
    "antidiagonal_set",
    "complement_set",
    "inverse_set",
    "lex_min",
    "reverse_set",
    "rotate_90_clockwise_set",
    "rotate_180_clockwise_set",
    "rotate_270_clockwise_set",
]



from permuta import Perm
from permuta.permutils import InsertionEncodablePerms


def test_is_insertion_encodable_rightmost():
    assert InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm((0, 1)),
            Perm((0, 1, 2)),
            Perm((4, 0, 2, 3, 5, 1)),
            Perm((0, 6, 3, 4, 2, 5, 1)),
            Perm((6, 0, 3, 7, 5, 1, 2, 4)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm(()),
            Perm((0, 1)),
            Perm((1, 0)),
            Perm((3, 0, 1, 2)),
            Perm((2, 1, 3, 0, 4)),
            Perm((4, 1, 3, 2, 5, 0)),
            Perm((5, 1, 2, 3, 4, 0)),
            Perm((2, 0, 4, 1, 5, 6, 3)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm((0,)),
            Perm((1, 0, 2, 3)),
            Perm((0, 4, 1, 2, 3)),
            Perm((6, 0, 1, 7, 4, 5, 2, 3)),
            Perm((7, 3, 6, 4, 2, 5, 1, 0)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm(()),
            Perm((1, 0)),
            Perm((2, 1, 0)),
            Perm((0, 3, 1, 2)),
            Perm((3, 2, 0, 5, 7, 4, 6, 1)),
            Perm((4, 1, 5, 7, 0, 3, 2, 6)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm(()),
            Perm((1, 0)),
            Perm((0, 2, 1)),
            Perm((1, 2, 0)),
            Perm((0, 1, 4, 2, 3)),
            Perm((2, 3, 0, 1, 4)),
            Perm((4, 7, 3, 0, 1, 5, 6, 2)),
            Perm((4, 7, 3, 2, 5, 6, 0, 1)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm((3, 2, 1, 4, 0)),
            Perm((4, 2, 3, 0, 1)),
            Perm((2, 3, 4, 0, 1, 5)),
            Perm((3, 2, 5, 4, 0, 1)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [Perm((4, 2, 5, 0, 3, 1)), Perm((6, 5, 3, 4, 1, 0, 2))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [Perm((2, 1, 0)), Perm((3, 1, 2, 0))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [Perm((6, 1, 3, 0, 2, 4, 5)), Perm((4, 3, 5, 6, 0, 2, 1, 7))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [Perm((1, 0, 2)), Perm((0, 2, 3, 1, 4))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm((0, 1, 2)),
            Perm((0, 2, 1)),
            Perm((1, 0, 2, 3)),
            Perm((3, 1, 0, 2)),
            Perm((3, 4, 1, 2, 0)),
            Perm((1, 5, 2, 3, 0, 4)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm((2, 0, 1)),
            Perm((0, 2, 3, 1)),
            Perm((3, 1, 0, 2)),
            Perm((3, 2, 1, 4, 0)),
            Perm((4, 0, 2, 3, 1)),
            Perm((4, 2, 0, 5, 1, 3)),
            Perm((6, 5, 0, 2, 7, 4, 3, 1)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm((1, 0, 2)),
            Perm((1, 2, 0)),
            Perm((2, 1, 0)),
            Perm((3, 6, 1, 5, 0, 2, 4)),
            Perm((6, 1, 2, 5, 4, 3, 0)),
            Perm((5, 4, 6, 0, 3, 2, 7, 1)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [Perm((1, 0, 2)), Perm((1, 3, 2, 0))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm((1, 2, 0)),
            Perm((2, 3, 0, 1)),
            Perm((3, 2, 1, 0)),
            Perm((1, 3, 2, 0, 4)),
            Perm((2, 1, 6, 0, 5, 4, 3)),
            Perm((4, 5, 0, 2, 3, 6, 7, 1)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm((2, 1, 0)),
            Perm((1, 2, 0, 3)),
            Perm((2, 0, 1, 3, 4)),
            Perm((4, 0, 1, 2, 3)),
            Perm((1, 3, 0, 5, 2, 4)),
            Perm((1, 5, 3, 0, 2, 4)),
            Perm((4, 3, 5, 0, 2, 1, 6)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm((1, 0, 2)),
            Perm((2, 3, 1, 0)),
            Perm((0, 4, 1, 3, 2)),
            Perm((6, 3, 5, 1, 2, 4, 0)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [Perm((0, 1, 2)), Perm((4, 2, 5, 3, 1, 0)), Perm((1, 2, 3, 0, 5, 6, 4))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [Perm((1, 2, 0)), Perm((2, 4, 0, 1, 3)), Perm((7, 5, 4, 3, 1, 0, 2, 6))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm((0, 1, 2)),
            Perm((1, 2, 0)),
            Perm((1, 0, 3, 2)),
            Perm((2, 0, 3, 1)),
            Perm((5, 3, 2, 4, 0, 1)),
            Perm((2, 0, 3, 4, 6, 5, 1)),
            Perm((5, 1, 6, 2, 7, 0, 4, 3)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm((0, 2, 1)),
            Perm((2, 0, 1)),
            Perm((2, 0, 3, 1, 4)),
            Perm((4, 0, 3, 1, 2)),
            Perm((4, 3, 1, 2, 0)),
            Perm((4, 2, 5, 1, 0, 3)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm((0, 2, 1)),
            Perm((1, 2, 0)),
            Perm((0, 3, 1, 2)),
            Perm((1, 0, 3, 2)),
            Perm((2, 4, 1, 0, 3)),
            Perm((4, 3, 0, 1, 2)),
            Perm((2, 5, 0, 3, 6, 4, 1)),
            Perm((5, 7, 1, 6, 3, 4, 0, 2)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm((0, 2, 1)),
            Perm((2, 0, 1, 3)),
            Perm((1, 3, 0, 2, 4)),
            Perm((5, 0, 4, 3, 2, 7, 6, 1)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm((0, 2, 1)),
            Perm((2, 1, 0)),
            Perm((3, 0, 2, 1)),
            Perm((0, 1, 2, 3, 4)),
            Perm((0, 1, 3, 2, 4)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm((0, 1, 2)),
            Perm((2, 0, 1, 3)),
            Perm((2, 1, 3, 0)),
            Perm((1, 0, 4, 2, 3)),
            Perm((1, 6, 2, 3, 4, 0, 5)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm((2, 0, 1)),
            Perm((3, 1, 0, 4, 2)),
            Perm((1, 6, 4, 2, 0, 3, 7, 5)),
            Perm((2, 6, 1, 5, 4, 7, 0, 3)),
            Perm((3, 0, 7, 2, 1, 4, 5, 6)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [Perm((1, 0, 2))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm((2, 0, 1)),
            Perm((3, 1, 0, 2)),
            Perm((1, 0, 5, 2, 4, 3)),
            Perm((4, 3, 0, 5, 6, 1, 2)),
            Perm((1, 2, 5, 6, 7, 3, 0, 4)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm((3, 4, 2, 1, 0)),
            Perm((3, 6, 2, 0, 5, 4, 1)),
            Perm((6, 3, 0, 2, 4, 5, 1)),
            Perm((5, 7, 0, 4, 2, 1, 3, 6)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm((3, 1, 0, 2)),
            Perm((4, 0, 2, 3, 1)),
            Perm((2, 5, 4, 1, 0, 3)),
            Perm((0, 6, 1, 2, 4, 5, 3)),
            Perm((6, 2, 4, 1, 5, 7, 3, 0)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm((1, 2, 0)),
            Perm((1, 2, 0, 3)),
            Perm((1, 5, 7, 2, 3, 4, 6, 0)),
            Perm((3, 6, 7, 4, 5, 2, 0, 1)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [Perm((2, 5, 1, 3, 0, 4))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [Perm((4, 0, 3, 1, 2)), Perm((1, 3, 5, 2, 0, 4)), Perm((1, 5, 4, 0, 3, 2))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [Perm((1, 2, 0)), Perm((4, 1, 2, 5, 0, 3)), Perm((1, 4, 0, 6, 2, 3, 7, 5))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm((1, 2, 0)),
            Perm((3, 4, 1, 5, 2, 0)),
            Perm((1, 4, 3, 5, 6, 2, 0)),
            Perm((3, 2, 4, 1, 6, 5, 0)),
            Perm((3, 0, 4, 2, 6, 7, 5, 1)),
            Perm((4, 1, 3, 6, 5, 2, 7, 0)),
            Perm((4, 2, 5, 0, 6, 1, 3, 7)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm((2, 1, 3, 0)),
            Perm((2, 1, 3, 0, 4)),
            Perm((4, 0, 3, 1, 2)),
            Perm((1, 3, 4, 2, 5, 0)),
            Perm((2, 0, 6, 4, 1, 3, 5)),
            Perm((3, 4, 1, 6, 2, 5, 7, 0)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [Perm((1, 3, 2, 0)), Perm((0, 4, 2, 1, 3)), Perm((4, 0, 3, 2, 1))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [
            Perm((2, 1, 3, 0)),
            Perm((3, 1, 2, 0)),
            Perm((1, 2, 0, 4, 3)),
            Perm((4, 2, 1, 0, 3)),
            Perm((0, 1, 2, 5, 3, 4)),
            Perm((3, 2, 5, 4, 0, 1)),
            Perm((2, 5, 7, 3, 4, 0, 6, 1)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [Perm((3, 5, 0, 1, 4, 2))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_rightmost(
        [Perm((0, 1, 2)), Perm((6, 1, 3, 0, 4, 2, 5, 7))]
    )


def test_is_insertion_encodable_maximum():
    assert InsertionEncodablePerms.is_insertion_encodable_maximum(
        [Perm((0,)), Perm((0, 1)), Perm((2, 0, 1, 3)), Perm((2, 1, 3, 0))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_maximum(
        [
            Perm((0, 1, 2)),
            Perm((1, 2, 0)),
            Perm((0, 1, 2, 4, 3)),
            Perm((2, 0, 4, 1, 3)),
            Perm((3, 0, 6, 1, 4, 2, 5)),
            Perm((3, 5, 1, 4, 6, 0, 2)),
            Perm((6, 1, 2, 3, 0, 5, 7, 4)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_maximum(
        [Perm((0, 2, 1)), Perm((2, 1, 0)), Perm((3, 1, 2, 0, 4, 5))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_maximum(
        [
            Perm((1, 2, 0)),
            Perm((2, 0, 1)),
            Perm((2, 3, 1, 0)),
            Perm((3, 1, 4, 0, 2)),
            Perm((0, 5, 4, 1, 3, 2)),
            Perm((3, 2, 5, 4, 0, 1)),
            Perm((0, 5, 3, 6, 4, 1, 7, 2)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_maximum(
        [
            Perm((1, 2, 0)),
            Perm((2, 1, 0)),
            Perm((1, 2, 0, 3)),
            Perm((2, 3, 1, 0, 4)),
            Perm((1, 2, 5, 6, 3, 4, 0)),
            Perm((5, 4, 6, 1, 0, 3, 2)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_maximum(
        [Perm((2, 1, 0, 3)), Perm((3, 1, 0, 2)), Perm((0, 1, 4, 2, 3))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_maximum(
        [
            Perm((0, 1, 2)),
            Perm((0, 2, 1)),
            Perm((2, 0, 1)),
            Perm((2, 1, 0)),
            Perm((1, 3, 2, 0)),
            Perm((1, 4, 3, 2, 0)),
            Perm((1, 2, 4, 0, 3, 6, 5)),
            Perm((1, 7, 0, 3, 4, 5, 6, 2)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_maximum(
        [
            Perm((2, 0, 1)),
            Perm((1, 2, 3, 0)),
            Perm((3, 0, 2, 1, 4)),
            Perm((4, 1, 6, 3, 0, 5, 2)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_maximum(
        [Perm((1, 0)), Perm((0, 2, 1)), Perm((1, 2, 0)), Perm((2, 1, 0))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_maximum(
        [Perm(()), Perm((0, 1, 3, 2))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_maximum(
        [
            Perm((0, 2, 1)),
            Perm((0, 1, 2, 3)),
            Perm((0, 3, 5, 1, 4, 2)),
            Perm((3, 5, 2, 6, 0, 1, 4, 7)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_maximum(
        [
            Perm((2, 0, 1, 3)),
            Perm((3, 1, 2, 0)),
            Perm((3, 2, 1, 0)),
            Perm((2, 3, 4, 1, 0)),
            Perm((4, 2, 0, 3, 5, 1)),
            Perm((6, 5, 4, 3, 7, 1, 2, 0)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_maximum(
        [Perm((0, 1, 2, 3)), Perm((3, 2, 5, 1, 4, 0))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_maximum(
        [Perm((0, 1)), Perm((0, 1, 2)), Perm((2, 1, 4, 0, 3)), Perm((3, 2, 0, 1, 4))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_maximum(
        [
            Perm((0, 1, 2)),
            Perm((0, 2, 1)),
            Perm((0, 2, 1, 4, 3)),
            Perm((2, 1, 4, 3, 0)),
            Perm((2, 4, 3, 5, 1, 0)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_maximum(
        [
            Perm((0, 1, 2)),
            Perm((2, 0, 1)),
            Perm((0, 4, 2, 1, 3)),
            Perm((1, 0, 3, 4, 2)),
            Perm((3, 2, 4, 0, 1)),
            Perm((1, 4, 5, 3, 0, 2)),
            Perm((0, 4, 2, 5, 1, 3, 6)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_maximum(
        [
            Perm((0, 2, 1)),
            Perm((0, 3, 2, 1)),
            Perm((1, 2, 3, 0)),
            Perm((1, 2, 3, 4, 0)),
            Perm((1, 4, 0, 3, 2)),
            Perm((4, 0, 1, 5, 2, 3)),
            Perm((3, 6, 1, 2, 0, 5, 4)),
            Perm((4, 0, 5, 1, 6, 2, 3)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_maximum(
        [
            Perm((0, 2, 1)),
            Perm((1, 2, 0, 3)),
            Perm((1, 3, 4, 0, 2)),
            Perm((1, 2, 5, 4, 3, 0)),
            Perm((5, 0, 2, 1, 4, 3)),
            Perm((4, 5, 2, 6, 3, 1, 0)),
            Perm((5, 0, 1, 4, 6, 3, 2)),
            Perm((5, 0, 3, 1, 6, 2, 4)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_maximum(
        [
            Perm((0, 2, 1)),
            Perm((1, 2, 0)),
            Perm((3, 1, 2, 0)),
            Perm((3, 2, 4, 0, 1)),
            Perm((2, 0, 5, 1, 3, 4)),
            Perm((3, 5, 0, 1, 4, 2, 7, 6)),
            Perm((4, 5, 2, 0, 3, 6, 1, 7)),
            Perm((6, 3, 2, 0, 1, 7, 4, 5)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable_maximum(
        [
            Perm((0, 3, 2, 1)),
            Perm((1, 2, 3, 0)),
            Perm((0, 2, 1, 4, 3)),
            Perm((6, 1, 0, 5, 3, 4, 2)),
            Perm((6, 0, 2, 5, 1, 4, 7, 3)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_maximum(
        [Perm((3, 0, 5, 2, 4, 1))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_maximum(
        [
            Perm((1, 0, 2)),
            Perm((1, 3, 2, 0)),
            Perm((1, 5, 4, 2, 0, 3)),
            Perm((4, 2, 0, 3, 5, 1)),
            Perm((4, 3, 2, 0, 5, 1)),
            Perm((5, 1, 2, 4, 3, 0, 6)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_maximum(
        [
            Perm((2, 0, 1)),
            Perm((1, 2, 0, 4, 3)),
            Perm((3, 0, 2, 1, 4, 5)),
            Perm((5, 3, 2, 0, 1, 6, 4)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_maximum(
        [Perm((0, 4, 3, 1, 2))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_maximum(
        [Perm((0, 2, 1)), Perm((0, 4, 2, 3, 1))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_maximum(
        [Perm((2, 3, 1, 5, 0, 4)), Perm((6, 2, 4, 1, 3, 7, 5, 0))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_maximum(
        [
            Perm((0, 4, 2, 1, 3)),
            Perm((1, 4, 0, 2, 3)),
            Perm((3, 1, 2, 5, 6, 0, 4)),
            Perm((0, 4, 6, 5, 7, 1, 2, 3)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_maximum(
        [Perm((1, 0, 3, 4, 2))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_maximum(
        [Perm((2, 0, 1, 3))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_maximum(
        [Perm((1, 3, 2, 4, 0))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_maximum(
        [
            Perm((0, 2, 1, 3)),
            Perm((0, 2, 3, 1)),
            Perm((0, 5, 3, 1, 2, 6, 4)),
            Perm((2, 4, 3, 0, 1, 6, 7, 5)),
            Perm((2, 7, 1, 4, 5, 6, 0, 3)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_maximum(
        [
            Perm((3, 0, 2, 1)),
            Perm((3, 1, 2, 0)),
            Perm((3, 1, 0, 4, 2)),
            Perm((4, 0, 2, 3, 1)),
            Perm((4, 0, 3, 1, 2)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_maximum(
        [Perm((1, 5, 6, 4, 2, 0, 3))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_maximum(
        [
            Perm((3, 2, 1, 0, 4)),
            Perm((4, 3, 2, 0, 5, 1)),
            Perm((5, 4, 2, 3, 0, 1)),
            Perm((0, 3, 2, 1, 4, 6, 5)),
            Perm((2, 3, 0, 5, 1, 6, 4)),
            Perm((3, 6, 2, 4, 1, 0, 5)),
            Perm((0, 5, 3, 4, 2, 6, 7, 1)),
            Perm((0, 6, 4, 3, 7, 1, 5, 2)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_maximum(
        [Perm((1, 2, 0)), Perm((4, 2, 5, 0, 1, 3)), Perm((6, 3, 5, 1, 7, 4, 0, 2))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_maximum(
        [Perm((3, 0, 2, 1)), Perm((4, 2, 1, 3, 0))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_maximum(
        [Perm((3, 2, 0, 1)), Perm((6, 1, 3, 2, 4, 0, 5))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_maximum(
        [
            Perm((0, 2, 3, 1)),
            Perm((0, 3, 2, 1)),
            Perm((3, 0, 1, 2)),
            Perm((4, 3, 2, 0, 5, 1)),
            Perm((1, 0, 6, 4, 3, 5, 2)),
            Perm((3, 1, 6, 2, 5, 0, 4)),
            Perm((5, 2, 6, 7, 0, 3, 1, 4)),
            Perm((7, 1, 6, 2, 5, 0, 4, 3)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_maximum(
        [Perm((0, 2, 1)), Perm((3, 1, 2, 0))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable_maximum(
        [
            Perm((0, 2, 1)),
            Perm((4, 0, 2, 1, 3)),
            Perm((2, 5, 4, 1, 3, 0)),
            Perm((3, 5, 4, 0, 2, 1)),
            Perm((1, 0, 4, 2, 3, 6, 5)),
            Perm((6, 3, 2, 0, 5, 1, 4)),
            Perm((4, 0, 1, 2, 3, 5, 6, 7)),
        ]
    )


def test_is_insertion_encodable():
    assert InsertionEncodablePerms.is_insertion_encodable(
        [Perm(()), Perm((0, 1)), Perm((0, 1, 2))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((0, 3, 2, 1)),
            Perm((0, 2, 3, 4, 1)),
            Perm((3, 0, 1, 2, 4)),
            Perm((4, 2, 1, 0, 3, 5)),
            Perm((0, 2, 1, 4, 6, 3, 5)),
            Perm((6, 2, 1, 0, 5, 4, 3)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [Perm((0, 1)), Perm((1, 0)), Perm((1, 0, 3, 2))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable([Perm(()), Perm((1, 0))])
    assert InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((0, 1, 2)),
            Perm((2, 0, 1)),
            Perm((0, 2, 4, 3, 1)),
            Perm((1, 2, 4, 0, 3)),
            Perm((3, 2, 4, 1, 5, 0)),
            Perm((2, 0, 3, 6, 4, 1, 5)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [Perm((1, 0)), Perm((0, 1, 2)), Perm((0, 3, 1, 2)), Perm((1, 3, 2, 0))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((2, 1, 0)),
            Perm((0, 3, 1, 2)),
            Perm((0, 3, 2, 1)),
            Perm((3, 2, 1, 0)),
            Perm((2, 1, 3, 5, 0, 4)),
            Perm((3, 1, 5, 2, 4, 0)),
            Perm((5, 1, 4, 3, 0, 2, 6)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [Perm(()), Perm((1, 0)), Perm((2, 0, 1)), Perm((0, 1, 2, 4, 3))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [Perm(()), Perm((0,)), Perm((0, 1)), Perm((1, 0)), Perm((0, 3, 1, 2, 4))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((0, 2, 1)),
            Perm((3, 2, 0, 1)),
            Perm((2, 1, 0, 3, 4)),
            Perm((3, 4, 2, 0, 5, 1)),
            Perm((4, 2, 7, 5, 0, 6, 3, 1)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable([Perm(()), Perm((2, 1, 0))])
    assert InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((2, 0, 1)),
            Perm((3, 0, 1, 2)),
            Perm((1, 0, 2, 3, 4)),
            Perm((2, 3, 0, 1, 4)),
            Perm((3, 4, 1, 0, 2)),
            Perm((4, 3, 1, 6, 0, 7, 2, 5)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((0, 1, 2)),
            Perm((1, 2, 0)),
            Perm((0, 2, 1, 3)),
            Perm((0, 3, 2, 1)),
            Perm((3, 0, 1, 2)),
            Perm((1, 0, 2, 4, 5, 6, 3)),
            Perm((1, 3, 6, 7, 2, 5, 4, 0)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [Perm(()), Perm((0,)), Perm((1, 0)), Perm((4, 0, 1, 2, 3))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((0, 1, 2)),
            Perm((2, 1, 0)),
            Perm((1, 2, 0, 3, 4)),
            Perm((3, 1, 0, 4, 2)),
            Perm((0, 5, 1, 4, 3, 2, 6)),
            Perm((2, 5, 4, 3, 6, 0, 1)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((1, 0, 2)),
            Perm((2, 3, 1, 0)),
            Perm((0, 1, 4, 3, 2)),
            Perm((0, 4, 3, 1, 2)),
            Perm((4, 0, 1, 3, 2)),
            Perm((0, 3, 2, 1, 5, 4)),
            Perm((2, 4, 0, 3, 5, 1)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((2, 0, 1)),
            Perm((3, 0, 1, 2)),
            Perm((0, 1, 4, 3, 2)),
            Perm((0, 5, 4, 7, 6, 3, 1, 2)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [Perm(()), Perm((1, 0, 2)), Perm((2, 0, 1, 3)), Perm((0, 3, 1, 2, 4))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((1, 2, 0)),
            Perm((2, 0, 1)),
            Perm((2, 1, 0)),
            Perm((3, 0, 2, 1)),
            Perm((1, 4, 2, 3, 5, 0)),
            Perm((1, 5, 0, 2, 3, 7, 6, 4)),
            Perm((4, 6, 5, 0, 2, 1, 3, 7)),
            Perm((7, 0, 1, 6, 3, 5, 4, 2)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((0, 2, 1)),
            Perm((2, 0, 1, 3)),
            Perm((0, 3, 2, 4, 1)),
            Perm((3, 4, 2, 0, 1)),
            Perm((1, 4, 5, 2, 0, 3)),
            Perm((4, 5, 3, 2, 0, 1)),
            Perm((6, 3, 0, 1, 2, 5, 4)),
            Perm((4, 3, 7, 5, 2, 1, 6, 0)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable([Perm((0,))])
    assert InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((0, 1, 2)),
            Perm((0, 2, 1)),
            Perm((2, 0, 1)),
            Perm((2, 3, 1, 0, 4)),
            Perm((4, 1, 3, 0, 2)),
            Perm((1, 3, 2, 0, 6, 4, 5)),
            Perm((1, 4, 5, 3, 2, 6, 0)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((3, 2, 1, 0)),
            Perm((1, 0, 2, 3, 4)),
            Perm((3, 1, 4, 0, 2, 7, 5, 6)),
            Perm((4, 5, 1, 7, 0, 6, 3, 2)),
            Perm((6, 5, 2, 4, 3, 0, 1, 7)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [Perm((0,)), Perm((0, 1)), Perm((1, 2, 0)), Perm((2, 1, 0))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [Perm((2, 0, 1)), Perm((2, 1, 3, 0, 4))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((1, 0, 2)),
            Perm((0, 2, 3, 1)),
            Perm((2, 1, 3, 0)),
            Perm((3, 0, 4, 1, 2)),
            Perm((6, 5, 4, 2, 0, 3, 1)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((0, 2, 1)),
            Perm((2, 0, 1)),
            Perm((2, 5, 4, 1, 0, 3)),
            Perm((3, 2, 1, 5, 0, 4)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((1, 0, 2, 3)),
            Perm((1, 3, 2, 0)),
            Perm((2, 1, 3, 0, 4)),
            Perm((5, 2, 1, 4, 3, 0)),
            Perm((0, 5, 1, 3, 6, 4, 2)),
            Perm((1, 4, 0, 2, 3, 5, 6)),
            Perm((6, 1, 2, 0, 3, 5, 4)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [Perm(()), Perm((1, 0)), Perm((1, 2, 0, 3)), Perm((3, 1, 0, 2))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((0, 2, 1)),
            Perm((1, 0, 2)),
            Perm((1, 2, 0)),
            Perm((3, 1, 2, 5, 4, 0)),
            Perm((0, 2, 3, 5, 6, 4, 1)),
            Perm((4, 5, 6, 3, 1, 0, 2)),
            Perm((0, 4, 6, 7, 2, 3, 5, 1)),
            Perm((7, 5, 3, 6, 2, 4, 0, 1)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [Perm((0, 1)), Perm((0, 1, 2)), Perm((1, 3, 0, 2)), Perm((0, 3, 4, 2, 1))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((1, 0, 2)),
            Perm((2, 0, 1)),
            Perm((1, 3, 4, 0, 2)),
            Perm((2, 6, 4, 5, 0, 1, 3)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((1, 2, 0)),
            Perm((2, 1, 0)),
            Perm((0, 2, 3, 1)),
            Perm((0, 2, 3, 1, 4)),
            Perm((0, 5, 3, 2, 4, 1)),
            Perm((0, 4, 6, 1, 5, 3, 2)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((1, 2, 0)),
            Perm((0, 2, 1, 3)),
            Perm((0, 3, 2, 1)),
            Perm((1, 0, 2, 3)),
            Perm((3, 0, 1, 2, 4)),
            Perm((4, 0, 1, 3, 2)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [Perm(()), Perm((3, 1, 2, 0)), Perm((1, 0, 3, 2, 4)), Perm((4, 2, 1, 0, 3))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [Perm((1, 2, 3, 0)), Perm((2, 1, 0, 3)), Perm((2, 1, 3, 0))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((1, 2, 0, 3)),
            Perm((1, 2, 3, 0)),
            Perm((4, 3, 0, 2, 1)),
            Perm((2, 1, 4, 7, 6, 3, 0, 5)),
            Perm((7, 4, 1, 3, 6, 5, 2, 0)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((0, 2, 1)),
            Perm((2, 3, 1, 0)),
            Perm((3, 1, 0, 2)),
            Perm((3, 1, 2, 0, 4)),
            Perm((2, 3, 1, 0, 4, 5)),
            Perm((1, 4, 7, 2, 6, 0, 5, 3)),
            Perm((5, 0, 3, 1, 7, 2, 4, 6)),
        ]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [Perm(()), Perm((0, 1, 3, 2)), Perm((2, 1, 4, 3, 0))]
    )
    assert InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((0, 1, 2)),
            Perm((1, 0, 3, 2)),
            Perm((0, 1, 4, 3, 2)),
            Perm((1, 4, 3, 0, 2)),
            Perm((4, 2, 1, 3, 0, 5)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [Perm((3, 1, 6, 2, 7, 0, 4, 5))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((0, 3, 1, 2)),
            Perm((1, 0, 3, 2)),
            Perm((1, 3, 2, 0, 4)),
            Perm((3, 2, 4, 0, 1)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((1, 2, 0)),
            Perm((0, 2, 3, 1)),
            Perm((2, 3, 1, 4, 0)),
            Perm((2, 3, 4, 0, 1)),
            Perm((4, 3, 1, 5, 0, 2)),
            Perm((0, 3, 2, 6, 4, 1, 5)),
            Perm((1, 5, 3, 2, 7, 6, 0, 4)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((2, 0, 1, 3)),
            Perm((2, 1, 0, 3)),
            Perm((3, 1, 0, 2)),
            Perm((3, 1, 2, 0)),
            Perm((6, 5, 1, 4, 0, 3, 2)),
            Perm((4, 2, 1, 0, 6, 3, 7, 5)),
            Perm((6, 2, 5, 3, 4, 1, 0, 7)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable([Perm((0, 3, 1, 2))])
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((0, 2, 1, 3)),
            Perm((1, 2, 3, 0, 4)),
            Perm((3, 1, 0, 2, 4)),
            Perm((3, 1, 4, 2, 0)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [Perm((1, 0, 2)), Perm((3, 1, 2, 6, 4, 0, 5)), Perm((3, 4, 2, 1, 0, 5, 6))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((0, 2, 1, 3)),
            Perm((2, 0, 3, 4, 1)),
            Perm((3, 4, 0, 1, 2)),
            Perm((4, 0, 3, 1, 2)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((0, 4, 3, 2, 1)),
            Perm((1, 0, 2, 4, 3)),
            Perm((2, 3, 0, 1, 4)),
            Perm((3, 1, 4, 2, 0)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [Perm((4, 5, 0, 1, 2, 6, 3))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable([Perm((0, 3, 4, 2, 1))])
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [Perm((3, 5, 6, 0, 7, 4, 2, 1))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable([Perm((3, 1, 2, 4, 0))])
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [Perm((1, 0, 3, 2)), Perm((2, 3, 0, 1, 4, 6, 5)), Perm((5, 4, 2, 1, 3, 0, 6))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [Perm((0, 2, 1)), Perm((2, 7, 6, 5, 0, 4, 3, 1))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((2, 0, 1, 3)),
            Perm((3, 0, 2, 1)),
            Perm((1, 2, 4, 0, 3)),
            Perm((2, 3, 4, 0, 1)),
            Perm((3, 0, 2, 4, 1)),
            Perm((4, 2, 1, 0, 3)),
            Perm((4, 3, 7, 2, 0, 6, 1, 5)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [Perm((2, 1, 0)), Perm((1, 3, 4, 2, 0)), Perm((0, 2, 6, 5, 1, 4, 3))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((0, 4, 5, 1, 3, 2)),
            Perm((5, 0, 2, 4, 1, 3)),
            Perm((0, 2, 4, 3, 6, 5, 1)),
            Perm((3, 0, 4, 1, 2, 5, 7, 6)),
            Perm((7, 5, 1, 6, 3, 2, 4, 0)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable([Perm((0, 2, 1, 3))])
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [Perm((0, 1, 2)), Perm((0, 4, 5, 1, 3, 2)), Perm((0, 6, 3, 5, 4, 1, 2))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [Perm((0, 2, 1)), Perm((0, 2, 1, 3)), Perm((0, 2, 3, 1))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [Perm((1, 2, 0)), Perm((2, 0, 3, 1)), Perm((2, 1, 3, 0))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable([Perm((3, 2, 1, 0))])
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [Perm((0, 1, 2)), Perm((2, 3, 0, 4, 1))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable([Perm((0, 1, 2, 3))])
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [Perm((0, 3, 1, 2)), Perm((3, 1, 0, 2, 4))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [Perm((2, 1, 0)), Perm((0, 4, 3, 2, 1)), Perm((4, 3, 0, 2, 1))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable([Perm((2, 1, 3, 0))])
    assert not InsertionEncodablePerms.is_insertion_encodable([Perm((1, 0, 2))])
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [Perm((1, 3, 4, 0, 2, 6, 5)), Perm((4, 0, 3, 1, 6, 5, 2))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [Perm((1, 2, 0)), Perm((3, 1, 2, 0))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((1, 2, 0, 3)),
            Perm((4, 2, 0, 5, 1, 3)),
            Perm((5, 4, 3, 1, 0, 2)),
            Perm((6, 1, 5, 4, 0, 2, 3)),
            Perm((7, 5, 4, 2, 0, 1, 6, 3)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [Perm((1, 3, 0, 2)), Perm((4, 0, 3, 2, 1))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((2, 3, 4, 5, 1, 0)),
            Perm((1, 0, 5, 3, 4, 2, 6)),
            Perm((0, 3, 6, 4, 1, 7, 2, 5)),
            Perm((0, 6, 5, 3, 2, 1, 4, 7)),
            Perm((4, 5, 1, 6, 0, 7, 3, 2)),
            Perm((6, 3, 2, 5, 1, 4, 0, 7)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable([Perm((0, 1, 2))])
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [Perm((0, 1, 2)), Perm((0, 1, 2, 3, 4)), Perm((2, 5, 1, 0, 4, 6, 3))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [Perm((0, 3, 2, 1, 4)), Perm((4, 3, 1, 0, 2))]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((2, 0, 1)),
            Perm((4, 2, 0, 1, 5, 6, 3)),
            Perm((6, 7, 2, 4, 5, 1, 3, 0)),
            Perm((7, 4, 2, 1, 3, 0, 5, 6)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((0, 1, 2)),
            Perm((1, 3, 0, 2, 4)),
            Perm((0, 1, 5, 3, 4, 6, 2)),
            Perm((0, 2, 6, 1, 3, 4, 5)),
        ]
    )
    assert not InsertionEncodablePerms.is_insertion_encodable(
        [
            Perm((4, 2, 3, 1, 0)),
            Perm((0, 4, 6, 3, 5, 2, 1)),
            Perm((2, 3, 0, 6, 1, 5, 4)),
        ]
    )



from typing import Iterable, List, Type

from permuta import Perm

from .abstract_strategy import EnumerationStrategy
from .core_strategies import core_strategies
from .insertion_encodable import InsertionEncodingStrategy

fast_enumeration_strategies: List[Type[EnumerationStrategy]] = [
    InsertionEncodingStrategy
]
fast_enumeration_strategies.extend(core_strategies)

long_enumeration_strategies: List[Type[EnumerationStrategy]] = []

all_enumeration_strategies: List[Type[EnumerationStrategy]] = (
    fast_enumeration_strategies + long_enumeration_strategies
)


def find_strategies(
    basis: Iterable[Perm], long_runnning: bool = True
) -> List[EnumerationStrategy]:
    """Test all enumeration strategies against the basis and return a list of
    potentially useful strategies. If `long_runnning` is False, test only the
    strategies that can be tested quickly.
    """
    if long_runnning:
        strategies: List[Type[EnumerationStrategy]] = all_enumeration_strategies
    else:
        strategies = fast_enumeration_strategies
    working_strategies: List[EnumerationStrategy] = []
    for strategy in strategies:
        strategy_object = strategy(basis)
        if strategy_object.applies():
            working_strategies.append(strategy_object)
    return working_strategies



from .patterns import BivincularPatt, CovincularPatt, MeshPatt, Perm, VincularPatt
from .perm_sets.permset import Av, Basis, MeshBasis

__all__ = [
    "Perm",
    "Av",
    "Basis",
    "MeshBasis",
    "MeshPatt",
    "BivincularPatt",
    "CovincularPatt",
    "VincularPatt",
]



from abc import abstractmethod, abstractproperty
from typing import ClassVar, FrozenSet, List, Type

from permuta import Av, MeshPatt, Perm
from permuta.enumeration_strategies.abstract_strategy import (
    EnumerationStrategyWithSymmetry,
)


class CoreStrategy(EnumerationStrategyWithSymmetry):
    """Abstract class for a core related strategy."""

    # https://arxiv.org/pdf/1912.07503.pdf
    # See this paper for corr_number

    @abstractproperty
    def patterns_needed(self) -> FrozenSet[Perm]:
        """Return the set of patterns that are needed for the strategy to be useful."""

    @staticmethod
    @abstractmethod
    def is_valid_extension(patt: Perm) -> bool:
        """Determine if the pattern satisfies the condition for strategy to apply."""

    def _applies_to_symmetry(self, basis: FrozenSet[Perm]):
        """Check if the core strategy applies to the basis or any of its symmetry."""
        assert isinstance(basis, frozenset)
        perm_class: Av = Av.from_iterable(basis)
        patterns_are_contained = all(p not in perm_class for p in self.patterns_needed)
        extensions_are_valid = all(
            self.is_valid_extension(patt)
            for patt in basis.difference(self.patterns_needed)
        )
        return patterns_are_contained and extensions_are_valid

    @classmethod
    def reference(cls) -> str:
        return (
            "Enumeration of Permutation Classes and Weighted Labelled "
            f"Independent Sets: Corollary {cls.corr_number}"
        )

    @property
    @staticmethod
    def corr_number() -> str:
        """The number of the corollary in the that gives this strategy."""
        raise NotImplementedError


def fstrip(perm: Perm) -> Perm:
    """Remove the leading 1 if the permutation is the sum of 1 + p."""
    assert len(perm) > 0
    if perm[0] == 0:
        return Perm.one_based(perm[1:])
    return perm


def bstrip(perm: Perm) -> Perm:
    """Remove the trailing n if the permutation is the sum of p + 1."""
    assert len(perm) > 0
    if perm[-1] == len(perm) - 1:
        return Perm(perm[:-1])
    return perm


def zero_plus_skewind(perm: Perm) -> bool:
    """Return True if the permutation is of the form 1 + p where p is a
    skew-indecomposable permutations
    """
    assert len(perm) > 0
    return perm[0] == 0 and not fstrip(perm).skew_decomposable()


def zero_plus_sumind(perm: Perm) -> bool:
    """Return True if the permutation is of the form 1 + p where p is a
    sum-indecomposable permutations
    """
    assert len(perm) > 0
    return perm[0] == 0 and not fstrip(perm).sum_decomposable()


def zero_plus_perm(perm: Perm) -> bool:
    """Return True if the permutation starts with a zero."""
    assert len(perm) > 0
    return perm[0] == 0


def last_sum_component(perm: Perm) -> Perm:
    """Return the last sum component of a permutation."""
    assert len(perm) > 0
    n, i = len(perm), 1
    comp = {perm[-1]}
    while comp != set(range(n - i, n)):
        i += 1
        comp.add(perm[n - i])
    return Perm.to_standard(perm[n - i : n])


def last_skew_component(perm: Perm) -> Perm:
    """Return the last skew component of a permutation."""
    assert len(perm) > 0
    n, i = len(perm), 1
    i = 1
    comp = {perm[-1]}
    while comp != set(range(i)):
        i += 1
        comp.add(perm[n - i])
    return Perm.to_standard(perm[n - i : n])


R_U: Perm = Perm((1, 2, 0, 3))  # 2314, row up
C_U: Perm = Perm((2, 0, 1, 3))  # 3124, colmn up
R_D: Perm = Perm((1, 3, 0, 2))  # 2413, row down
C_D: Perm = Perm((2, 0, 3, 1))  # 3142, column down


class RuCuCoreStrategy(CoreStrategy):
    """This strategies uses independent set of the up-core graph to enumerate a
    class as inflation of an independent set.
    """

    patterns_needed: FrozenSet[Perm] = frozenset([R_U, C_U])
    corr_number: ClassVar[str] = "4.3"

    @staticmethod
    def is_valid_extension(patt: Perm) -> bool:
        return zero_plus_skewind(patt)


class RdCdCoreStrategy(CoreStrategy):
    """This strategies uses independent set of the down-core graph to enumerate a
    class as inflation of an independent set.
    """

    patterns_needed = frozenset([R_D, C_D])
    corr_number: ClassVar[str] = "4.6"

    @staticmethod
    def is_valid_extension(patt: Perm) -> bool:
        return zero_plus_sumind(patt)


class RuCuRdCdCoreStrategy(CoreStrategy):
    """TODO"""

    patterns_needed = frozenset([R_D, C_D, R_U, C_U])
    corr_number: ClassVar[str] = "5.4"

    @staticmethod
    def is_valid_extension(patt: Perm) -> bool:
        return zero_plus_perm(patt)


class RuCuCdCoreStrategy(CoreStrategy):
    """TODO"""

    patterns_needed = frozenset([R_U, C_U, C_D])
    corr_number: ClassVar[str] = "6.3"

    @staticmethod
    def is_valid_extension(patt: Perm) -> bool:
        return zero_plus_skewind(patt)


class RdCdCuCoreStrategy(CoreStrategy):
    """TODO"""

    patterns_needed = frozenset([R_D, C_D, C_U])
    corr_number: ClassVar[str] = "7.4"

    @staticmethod
    def is_valid_extension(patt):
        return zero_plus_sumind(bstrip(patt))


class RdCuCoreStrategy(CoreStrategy):
    """TODO"""

    patterns_needed = frozenset([R_D, C_U])
    corr_number: ClassVar[str] = "8.3"

    @staticmethod
    def is_valid_extension(patt):
        return zero_plus_skewind(patt) and zero_plus_sumind(bstrip(patt))


class Rd2134CoreStrategy(CoreStrategy):
    """TODO"""

    _NON_INC: ClassVar[Av] = Av.from_iterable([Perm((0, 1))])
    _M_PATT: ClassVar[MeshPatt] = MeshPatt(
        Perm((1, 0)), [(0, 1), (0, 2), (1, 0), (1, 1), (1, 2), (2, 1), (2, 2)]
    )

    patterns_needed = frozenset([R_D, Perm((1, 0, 2, 3))])
    corr_number: ClassVar[str] = "9.5"

    @staticmethod
    def is_valid_extension(patt: Perm) -> bool:
        last_comp = last_sum_component(fstrip(patt))
        return (
            patt[0] == 0
            and fstrip(patt).avoids(Rd2134CoreStrategy._M_PATT)
            and (last_comp not in Rd2134CoreStrategy._NON_INC or len(last_comp) == 1)
        )


class Ru2143CoreStrategy(CoreStrategy):
    """TODO"""

    _NON_DEC: ClassVar[Av] = Av.from_iterable([Perm((1, 0))])
    _M_PATT: ClassVar[MeshPatt] = MeshPatt(
        Perm((0, 1)), [(0, 1), (0, 2), (1, 0), (1, 1), (1, 2), (2, 1), (2, 2)]
    )

    patterns_needed = frozenset([R_U, Perm((1, 0, 3, 2))])
    corr_number: ClassVar[str] = "10.5"

    @staticmethod
    def is_valid_extension(patt: Perm) -> bool:
        patt = fstrip(patt)
        return (
            patt.avoids(Ru2143CoreStrategy._M_PATT)
            and last_skew_component(patt) not in Ru2143CoreStrategy._NON_DEC
        )


core_strategies: List[Type[CoreStrategy]] = [
    RuCuCoreStrategy,
    RdCdCoreStrategy,
    RuCuRdCdCoreStrategy,
    RuCuCdCoreStrategy,
    RdCdCuCoreStrategy,
    RdCuCoreStrategy,
    Rd2134CoreStrategy,
    Ru2143CoreStrategy,
]
